import os
from copy import deepcopy
from typing import Union

import numpy as np
import pandas as pd
from numba import float32, float64, njit, prange

from simba.feature_extractors.perimeter_jit import jitted_hull
from simba.mixins.circular_statistics import CircularStatisticsMixin
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


@njit(
    [
        (float32[:, :], float32[:, :], float64),
        (float64[:, :], float64[:, :], float64),
        (float64[:, :], float32[:, :], float64),
    ]
)
def calculate_weighted_avg(
    bp: np.ndarray, p: Union[np.ndarray, None], threshold: float
):
    results = np.full((bp.shape[0]), np.nan)
    n = bp.shape[0]
    for i in prange(n):
        if p is not None:
            p_thresh_idx = np.argwhere(p[i] > threshold).flatten()
            if p_thresh_idx.shape[0] > 0:
                p_vals = p[i][p_thresh_idx]
                bp_vals = bp[i][p_thresh_idx]
                weighted_sum = 0
                for x in range(p_vals.shape[0]):
                    weighted_sum += bp_vals[x] * p_vals[x]
                frm_result = weighted_sum / np.sum(p_vals)
            else:
                frm_result = np.mean(bp[i])
        else:
            frm_result = np.mean(bp[i])
        results[i] = frm_result
    return results


@njit([(float32[:, :], float32[:, :], float32[:, :], float64)])
def polygon_fill(
    x_bps: np.ndarray, y_bps: np.ndarray, p_bps: np.ndarray, threshold: float
):
    results = np.full((x_bps.shape[0], x_bps[0].shape[0], 2), np.nan)
    for frm_cnt in prange(x_bps.shape[0]):
        frm_bp_cnt = 0
        for bp_cnt in range(p_bps[frm_cnt].shape[0]):
            if p_bps[frm_cnt][bp_cnt] > threshold:
                results[frm_cnt][bp_cnt] = np.array(
                    [x_bps[frm_cnt, bp_cnt], y_bps[frm_cnt, bp_cnt]]
                )
                frm_bp_cnt += 1
        if frm_bp_cnt < 3:
            results[frm_cnt] = np.zeros((x_bps[0].shape[0], 2))
        elif np.sum(np.isnan(results[frm_cnt])) > 0:
            fill_idx = np.argwhere(~np.isnan(results[frm_cnt]))[0]
            fill_val = results[frm_cnt][fill_idx[0]]
            results[frm_cnt][np.isnan(results[frm_cnt][:, 0])] = fill_val[0]
            results[frm_cnt][np.isnan(results[frm_cnt][:, 1])] = fill_val[1]

    return results


def get_circle_fit_angle(x: np.ndarray, y: np.ndarray, p: np.ndarray, threshold: float):
    combined_arr = np.stack((x, y), axis=2).astype(np.float64)
    circles = CircularStatisticsMixin.fit_circle(data=combined_arr)
    diff_x_last = x[:, -1] - circles[:, 0]
    diff_y_last = y[:, -1] - circles[:, 1]
    diff_x_first = x[:, 0] - circles[:, 0]
    diff_y_first = y[:, 0] - circles[:, 1]
    angles = np.degrees(
        np.arctan2(diff_y_last, diff_x_last) - np.arctan2(diff_y_first, diff_x_first)
    )
    angles = angles + 360 * (angles < 0)

    below_thresh_idx = np.argwhere(np.average(p, axis=1) < threshold)
    angles[below_thresh_idx] = 0.0

    return angles


class AmberFeatureExtractor(ConfigReader, FeatureExtractionMixin):
    """
    Class for extracting features for the AMBER pipeline.

    The AMBER pipeline was developed for quantifying home-cage maternal and mother-pup interactions from side-view recordings.
    Downstream behavior classifiers can be trained to assess maternal nest attendance, nursing, pup-directed licking and grooming, self-directed grooming, eating, and drinking.

    .. note::
       `AMBER publication <https://www.nature.com/articles/s41598-023-45495-4>`_.
       `AMBER GitHub Repository <https://github.com/lapphe/AMBER-pipeline>`_.
       `AMBER OSF Repository <https://osf.io/e3dyc/>`_.

       For more info, contact Hannah E. Lapp `Hannah.Lapp@austin.utexas.edu <Hannah.Lapp@austin.utexas.edu/>`_.

    :example:
    >>> AmberFeatureExtractor(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')

    References
    ----------
       .. [1] Lapp, Hannah E., Melissa G. Salazar, and Frances A. Champagne. “Automated Maternal Behavior during Early Life in Rodents (AMBER) Pipeline.” Scientific Reports 13, no. 1 (October 25, 2023): 18277. https://doi.org/10.1038/s41598-023-45495-4.
    """

    def __init__(self, config_path: str):
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        self.timer = SimbaTimer(start=True)
        print(f"Extracting features from {len(self.files_found)} file(s)...")
        self.pup_threshold = 0.4
        self.dam_threshold = 0.4
        self.roll_windows_values = [
            1,
            2,
            5,
            8,
            0.5,
        ]  # values used to calculate rolling average across frames

    def run(self):
        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, file_name, _ = get_fn_ext(file_path)
            print(
                f"Extracting features for video {file_name} ({file_cnt + 1}/{len(self.files_found)})..."
            )
            video_settings, self.px_per_mm, fps = self.read_video_info(
                video_name=file_name
            )
            data_df = read_df(file_path=file_path, file_type=self.file_type)

            body_part_names = [
                x.replace("_x", "") for x in list(data_df.columns) if "_x" in x
            ]
            pup_bp_names = [x for x in body_part_names if "pup" in x]
            dam_bp_names = [x for x in body_part_names if x not in pup_bp_names]
            p_col_names = [p + "_p" for p in body_part_names]

            print("Calculating dam points, centroids, and convex hulls...")

            data_df["arm_x"] = np.where(
                data_df["left_armpit_p"] > self.dam_threshold,
                data_df["left_armpit_x"],
                data_df["right_armpit_x"],
            )
            data_df["arm_y"] = np.where(
                data_df["left_armpit_p"] > self.dam_threshold,
                data_df["left_armpit_y"],
                data_df["right_armpit_y"],
            )
            data_df["arm_p"] = np.where(
                data_df["left_armpit_p"] > self.dam_threshold,
                data_df["left_armpit_p"],
                data_df["right_armpit_p"],
            )

            data_df["side_x"] = np.where(
                data_df["left_ventrum_side_p"] > self.dam_threshold,
                data_df["left_ventrum_side_x"],
                data_df["right_ventrum_side_x"],
            )
            data_df["side_y"] = np.where(
                data_df["left_ventrum_side_p"] > self.dam_threshold,
                data_df["left_ventrum_side_y"],
                data_df["right_ventrum_side_y"],
            )
            data_df["side_p"] = np.where(
                data_df["left_ventrum_side_p"] > self.dam_threshold,
                data_df["left_ventrum_side_p"],
                data_df["right_ventrum_side_p"],
            )

            dam_x_bps, dam_y_bps, dam_p_bps = (
                [x + "_x" for x in dam_bp_names],
                [x + "_y" for x in dam_bp_names],
                [x + "_p" for x in dam_bp_names],
            )
            data_df["dam_centroid_x"] = calculate_weighted_avg(
                bp=data_df[dam_x_bps].values,
                p=data_df[dam_p_bps].values,
                threshold=self.dam_threshold,
            )
            data_df["dam_centroid_y"] = calculate_weighted_avg(
                bp=data_df[dam_y_bps].values,
                p=data_df[dam_p_bps].values,
                threshold=self.dam_threshold,
            )

            dam_polygon_points = polygon_fill(
                x_bps=data_df[dam_x_bps].values,
                y_bps=data_df[dam_y_bps].values,
                p_bps=data_df[dam_p_bps].values,
                threshold=0.2,
            )
            data_df["dam_convex_hull"] = jitted_hull(
                points=dam_polygon_points.astype(np.float32), target="area"
            ) / (self.px_per_mm**2)

            dam_head_parts = [
                "dam_nose",
                "left_eye",
                "right_eye",
                "left_ear",
                "right_ear",
                "top_head_dam",
            ]
            dam_head_x, dam_head_y, dam_head_p = (
                [x + "_x" for x in dam_head_parts],
                [x + "_y" for x in dam_head_parts],
                [x + "_p" for x in dam_head_parts],
            )

            data_df["head_centroid_x"] = calculate_weighted_avg(
                bp=data_df[dam_head_x].values,
                p=data_df[dam_head_p].values,
                threshold=self.dam_threshold,
            )
            data_df["head_centroid_y"] = calculate_weighted_avg(
                bp=data_df[dam_head_y].values,
                p=data_df[dam_head_p].values,
                threshold=self.dam_threshold,
            )

            head_polygon = polygon_fill(
                x_bps=data_df[dam_head_x].values,
                y_bps=data_df[dam_head_y].values,
                p_bps=data_df[dam_head_p].values,
                threshold=0.2,
            )

            data_df["head_convex_hull"] = jitted_hull(
                points=head_polygon.astype(np.float32), target="area"
            ) / (self.px_per_mm**2)

            print("Calculating pup centroids...")
            pup_x_bps, pup_y_bps, pup_p_bps = (
                [x + "_x" for x in pup_bp_names],
                [x + "_y" for x in pup_bp_names],
                [x + "_p" for x in pup_bp_names],
            )

            data_df["pups_centroid_x"] = calculate_weighted_avg(
                bp=data_df[pup_x_bps].values,
                p=data_df[pup_p_bps].values,
                threshold=self.pup_threshold,
            )
            data_df["pups_centroid_y"] = calculate_weighted_avg(
                bp=data_df[pup_y_bps].values,
                p=data_df[pup_p_bps].values,
                threshold=self.pup_threshold,
            )

            print("Calculating pup convex hull...")
            pups_polygon = polygon_fill(
                x_bps=data_df[pup_x_bps].values,
                y_bps=data_df[pup_y_bps].values,
                p_bps=data_df[pup_p_bps].values,
                threshold=0.2,
            )
            data_df["pups_convex_hull"] = jitted_hull(
                points=pups_polygon.astype(np.float32), target="area"
            ) / (self.px_per_mm**2)

            print("Calculating high probably body part counts...")
            data_df["pup_avg_p"] = data_df[pup_p_bps].mean(axis=1)

            data_df["high_p_pup_bp"] = FeatureExtractionMixin().count_values_in_range(
                data=data_df[pup_p_bps].fillna(0).values, ranges=np.array([[0.2, 1.0]])
            )
            data_df["high_p_dam_bp"] = FeatureExtractionMixin().count_values_in_range(
                data=data_df[dam_p_bps].fillna(0).values, ranges=np.array([[0.2, 1.0]])
            )

            data_df["high_p_pup_bp_no_zero"] = data_df["high_p_pup_bp"].replace(
                to_replace=0, value=1
            )

            data_df["pups_centroid_mult_x"] = (
                data_df["pups_centroid_x"] * data_df["high_p_pup_bp_no_zero"]
            )
            data_df["pups_centroid_mult_y"] = (
                data_df["pups_centroid_y"] * data_df["high_p_pup_bp_no_zero"]
            )

            roll_rows_30 = int(30 * 60 * fps)
            data_df["pups_centroid_x_roll_mean_30m"] = (
                data_df["pups_centroid_mult_x"]
                .rolling(roll_rows_30, min_periods=0, center=True)
                .sum()
                / data_df["high_p_pup_bp_no_zero"]
                .rolling(roll_rows_30, min_periods=0, center=True)
                .sum()
            )
            data_df["pups_centroid_y_roll_mean_30m"] = (
                data_df["pups_centroid_mult_y"]
                .rolling(roll_rows_30, min_periods=0, center=True)
                .sum()
                / data_df["high_p_pup_bp_no_zero"]
                .rolling(roll_rows_30, min_periods=0, center=True)
                .sum()
            )

            roll_rows_60 = int(60 * 60 * fps)
            data_df["pups_centroid_x_roll_mean_60m"] = (
                data_df["pups_centroid_mult_x"]
                .rolling(roll_rows_60, min_periods=0, center=True)
                .sum()
                / data_df["high_p_pup_bp_no_zero"]
                .rolling(roll_rows_60, min_periods=0, center=True)
                .sum()
            )
            data_df["pups_centroid_y_roll_mean_60m"] = (
                data_df["pups_centroid_mult_y"]
                .rolling(roll_rows_60, min_periods=0, center=True)
                .sum()
                / data_df["high_p_pup_bp_no_zero"]
                .rolling(roll_rows_60, min_periods=0, center=True)
                .sum()
            )

            print("Calculating movements...")

            movement_columns = deepcopy(dam_bp_names)
            data_df_shifted = FeatureExtractionMixin.create_shifted_df(
                df=data_df, periods=1
            )

            for col in movement_columns:
                bp_cols, s_bp_cols = [f"{col}_x", f"{col}_y"], [
                    f"{col}_x_shifted",
                    f"{col}_y_shifted",
                ]
                bp_arr, s_bp_arr = (
                    data_df_shifted[bp_cols].values,
                    data_df_shifted[s_bp_cols].values,
                )
                data_df[f"{col}_movement"] = (
                    FeatureExtractionMixin.framewise_euclidean_distance(
                        location_1=bp_arr, location_2=s_bp_arr, px_per_mm=self.px_per_mm
                    )
                )

            back_point_movements = [
                "back_2_movement",
                "back_3_movement",
                "back_4_movement",
                "back_5_movement",
                "back_6_movement",
                "back_7_movement",
                "back_8_movement",
                "back_9_movement",
                "back_10_movement",
            ]
            back_point_movements_p = [
                "back_2_p",
                "back_3_p",
                "back_4_p",
                "back_5_p",
                "back_6_p",
                "back_7_p",
                "back_8_p",
                "back_9_p",
                "back_10_p",
            ]

            data_df["back_avg_movement"] = calculate_weighted_avg(
                bp=data_df[back_point_movements].values,
                p=data_df[back_point_movements_p].values,
                threshold=self.dam_threshold,
            )

            head_point_movements = [
                "dam_nose_movement",
                "right_eye_movement",
                "left_eye_movement",
                "left_ear_movement",
                "right_ear_movement",
            ]
            head_point_movements_p = [
                "dam_nose_p",
                "right_eye_p",
                "left_eye_p",
                "left_ear_p",
                "right_ear_p",
            ]

            data_df["head_avg_movement"] = calculate_weighted_avg(
                bp=data_df[head_point_movements].values,
                p=data_df[head_point_movements_p].values,
                threshold=self.dam_threshold,
            )
            #
            data_df["head_max_movement"] = np.ma.max(
                [
                    data_df["dam_nose_movement"],
                    data_df["right_eye_movement"],
                    data_df["left_eye_movement"],
                    data_df["left_ear_movement"],
                    data_df["right_ear_movement"],
                ],
                axis=0,
            )

            data_df["ventrum_side_movement"] = calculate_weighted_avg(
                bp=data_df[
                    ["left_ventrum_side_movement", "right_ventrum_side_movement"]
                ].values,
                p=data_df[["left_ventrum_side_p", "right_ventrum_side_p"]].values,
                threshold=self.dam_threshold,
            )

            data_df["leg_front_movement"] = calculate_weighted_avg(
                bp=data_df[
                    ["left_leg_front_movement", "right_leg_front_movement"]
                ].values,
                p=data_df[["left_leg_front_p", "right_leg_front_p"]].values,
                threshold=self.dam_threshold,
            )

            data_df["leg_behind_movement"] = calculate_weighted_avg(
                bp=data_df[
                    ["left_leg_behind_movement", "right_leg_behind_movement"]
                ].values,
                p=data_df[["left_leg_behind_p", "right_leg_behind_p"]].values,
                threshold=self.dam_threshold,
            )

            data_df["wrist_movement"] = calculate_weighted_avg(
                bp=data_df[["left_wrist_movement", "right_wrist_movement"]].values,
                p=data_df[["left_wrist_p", "right_wrist_p"]].values,
                threshold=self.dam_threshold,
            )

            data_df["armpit_movement"] = calculate_weighted_avg(
                bp=data_df[["left_armpit_movement", "right_armpit_movement"]].values,
                p=data_df[["left_armpit_p", "right_armpit_p"]].values,
                threshold=self.dam_threshold,
            )

            data_df["shoulder_movement"] = calculate_weighted_avg(
                bp=data_df[
                    ["left_shoulder_movement", "right_shoulder_movement"]
                ].values,
                p=data_df[["left_shoulder_p", "right_shoulder_p"]].values,
                threshold=self.dam_threshold,
            )

            data_df["eye_movement"] = calculate_weighted_avg(
                bp=data_df[["left_eye_movement", "right_eye_movement"]].values,
                p=data_df[["left_eye_p", "right_eye_p"]].values,
                threshold=self.dam_threshold,
            )

            data_df["ear_movement"] = calculate_weighted_avg(
                bp=data_df[["left_ear_movement", "right_ear_movement"]].values,
                p=data_df[["left_ear_p", "right_ear_p"]].values,
                threshold=self.dam_threshold,
            )

            data_df["ankle_movement"] = calculate_weighted_avg(
                bp=data_df[["left_ankle_movement", "right_ankle_movement"]].values,
                p=data_df[["left_ankle_p", "right_ankle_p"]].values,
                threshold=self.dam_threshold,
            )

            print("Calculating distances...")

            data_df["dam_pup30m_distance"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["dam_centroid_x", "dam_centroid_y"]].values,
                    location_2=data_df[
                        [
                            "pups_centroid_x_roll_mean_30m",
                            "pups_centroid_y_roll_mean_30m",
                        ]
                    ].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["head_pup30m_distance"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["head_centroid_x", "head_centroid_y"]].values,
                    location_2=data_df[
                        [
                            "pups_centroid_x_roll_mean_30m",
                            "pups_centroid_y_roll_mean_30m",
                        ]
                    ].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["dam_pup60m_distance"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["dam_centroid_x", "dam_centroid_y"]].values,
                    location_2=data_df[
                        [
                            "pups_centroid_x_roll_mean_60m",
                            "pups_centroid_y_roll_mean_60m",
                        ]
                    ].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["head_pup60m_distance"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["head_centroid_x", "head_centroid_y"]].values,
                    location_2=data_df[
                        [
                            "pups_centroid_x_roll_mean_60m",
                            "pups_centroid_y_roll_mean_60m",
                        ]
                    ].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["dam_pup_distance"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["dam_centroid_x", "dam_centroid_y"]].values,
                    location_2=data_df[["pups_centroid_x", "pups_centroid_y"]].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["head_pup_distance"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["head_centroid_x", "head_centroid_y"]].values,
                    location_2=data_df[["pups_centroid_x", "pups_centroid_y"]].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["back_length"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["back_2_x", "back_2_y"]].values,
                    location_2=data_df[["back_10_x", "back_10_y"]].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["nose_back10_length"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["dam_nose_x", "dam_nose_y"]].values,
                    location_2=data_df[["back_10_x", "back_10_y"]].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["left_wrist_nose_length"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["left_wrist_x", "left_wrist_y"]].values,
                    location_2=data_df[["dam_nose_x", "dam_nose_y"]].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["right_wrist_nose_length"] = (
                FeatureExtractionMixin.framewise_euclidean_distance(
                    location_1=data_df[["right_wrist_x", "right_wrist_y"]].values,
                    location_2=data_df[["dam_nose_x", "dam_nose_y"]].values,
                    px_per_mm=self.px_per_mm,
                )
            )

            data_df["wrist_nose_length"] = calculate_weighted_avg(
                bp=data_df[
                    ["left_wrist_nose_length", "right_wrist_nose_length"]
                ].values,
                p=data_df[["left_wrist_p", "right_wrist_p"]].values,
                threshold=self.dam_threshold,
            )
            data_df.drop(
                inplace=True,
                columns=["left_wrist_nose_length", "right_wrist_nose_length"],
            )
            dam_p = [
                "dam_nose_p",
                "left_eye_p",
                "right_eye_p",
                "left_ear_p",
                "right_ear_p",
                "left_shoulder_p",
                "right_shoulder_p",
                "arm_p",
                "side_p",
            ]
            data_df["avg_dam_bp_p"] = np.mean(data_df[dam_p].values, axis=1)

            data_df["sum_probabilities"] = data_df[p_col_names].sum(axis=1)

            print("Calculating fields for dam back curve...")
            back_points_x = [
                "back_1_center_x",
                "back_2_x",
                "back_3_x",
                "back_4_x",
                "back_5_x",
                "back_6_x",
                "back_7_x",
                "back_8_x",
                "back_9_x",
                "back_10_x",
            ]
            back_points_y = [
                "back_1_center_y",
                "back_2_y",
                "back_3_y",
                "back_4_y",
                "back_5_y",
                "back_6_y",
                "back_7_y",
                "back_8_y",
                "back_9_y",
                "back_10_y",
            ]
            back_points_p = [
                "back_1_center_p",
                "back_2_p",
                "back_3_p",
                "back_4_p",
                "back_5_p",
                "back_6_p",
                "back_7_p",
                "back_8_p",
                "back_9_p",
                "back_10_p",
            ]

            print("Calculating dam angles...")

            data_df["dam_back_angle"] = get_circle_fit_angle(
                x=data_df[back_points_x].values,
                y=data_df[back_points_y].values,
                p=data_df[back_points_p].values,
                threshold=0.5,
            )

            head_angle_cords = [
                "back_10_x",
                "back_10_y",
                "back_1_center_x",
                "back_1_center_y",
                "top_head_dam_x",
                "top_head_dam_y",
            ]
            data_df["dam_head_angle"] = FeatureExtractionMixin.angle3pt_serialized(
                data=data_df[head_angle_cords].values
            )

            top_angle_cords = [
                "back_10_x",
                "back_10_y",
                "back_4_x",
                "back_4_y",
                "back_1_center_x",
                "back_1_center_y",
            ]
            data_df["dam_back_top_angle"] = FeatureExtractionMixin.angle3pt_serialized(
                data=data_df[top_angle_cords].values
            )

            head_angle_cords = [
                "back_2_x",
                "back_2_y",
                "back_1_center_x",
                "back_1_center_y",
                "top_head_dam_x",
                "top_head_dam_y",
            ]
            data_df["dam_head_angle2"] = FeatureExtractionMixin.angle3pt_serialized(
                data=data_df[head_angle_cords].values
            )

            print("Calculating rolling averages...")

            roll_windows = []
            for j in range(len(self.roll_windows_values)):
                roll_windows.append(int(fps / self.roll_windows_values[j]))

            data_df["head_avg_movement_roll_mean_1s"] = (
                data_df["head_avg_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )

            data_df["head_avg_movement_roll_mean_1ds"] = (
                data_df["head_avg_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["head_avg_movement_roll_mean_2s"] = (
                data_df["head_avg_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["back_avg_movement_roll_mean_1s"] = (
                data_df["back_avg_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["back_avg_movement_roll_mean_1ds"] = (
                data_df["back_avg_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["back_avg_movement_roll_mean_2s"] = (
                data_df["back_avg_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["head_back_rel_roll_mean_1s"] = data_df[
                "head_avg_movement_roll_mean_1s"
            ] / (
                data_df["head_avg_movement_roll_mean_1s"]
                + data_df["back_avg_movement_roll_mean_1s"]
            )
            data_df["head_back_rel_roll_mean_1ds"] = data_df[
                "head_avg_movement_roll_mean_1ds"
            ] / (
                data_df["head_avg_movement_roll_mean_1ds"]
                + data_df["back_avg_movement_roll_mean_1ds"]
            )
            data_df["head_back_rel_roll_mean_2s"] = data_df[
                "head_avg_movement_roll_mean_2s"
            ] / (
                data_df["head_avg_movement_roll_mean_2s"]
                + data_df["back_avg_movement_roll_mean_2s"]
            )

            data_df["pups_convex_hull_roll_mean_1s"] = (
                data_df["pups_convex_hull"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["pups_convex_hull_roll_mean_1ds"] = (
                data_df["pups_convex_hull"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["pups_convex_hull_roll_mean_2s"] = (
                data_df["pups_convex_hull"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_pup_distance_roll_mean_1s"] = (
                data_df["dam_pup_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_pup_distance_roll_mean_1ds"] = (
                data_df["dam_pup_distance"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_pup_distance_roll_mean_2s"] = (
                data_df["dam_pup_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_pup30m_distance_roll_sum_1s"] = (
                data_df["dam_pup30m_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_pup30m_distance_roll_mean_1ds"] = (
                data_df["dam_pup30m_distance"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_pup30m_distance_roll_mean_2s"] = (
                data_df["dam_pup30m_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_pup60m_distance_roll_mean_1s"] = (
                data_df["dam_pup60m_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_pup60m_distance_roll_mean_1ds"] = (
                data_df["dam_pup60m_distance"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_pup60m_distance_roll_mean_2s"] = (
                data_df["dam_pup60m_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_back_angle_roll_mean_1s"] = (
                data_df["dam_back_angle"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            # csv_df['dam_back_angle_roll_mean_1ds'] = csv_df['dam_back_angle'].rolling(roll_windows[2], min_periods=0, center=True).mean()
            data_df["dam_back_angle_roll_mean_2s"] = (
                data_df["dam_back_angle"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_head_angle_roll_mean_1s"] = (
                data_df["dam_head_angle"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_head_angle_roll_mean_1ds"] = (
                data_df["dam_head_angle"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_head_angle_roll_mean_2s"] = (
                data_df["dam_head_angle"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_head_angle2_roll_mean_1s"] = (
                data_df["dam_head_angle2"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_head_angle2_roll_mean_1ds"] = (
                data_df["dam_head_angle2"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_head_angle2_roll_mean_2s"] = (
                data_df["dam_head_angle2"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_back_top_angle_roll_mean_1s"] = (
                data_df["dam_back_top_angle"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            # csv_df['dam_back_top_angle_roll_mean_1ds'] = csv_df['dam_back_top_angle'].rolling(roll_windows[2], min_periods=0, center=True).mean()
            data_df["dam_back_top_angle_roll_mean_2s"] = (
                data_df["dam_back_top_angle"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_nose_movement_roll_mean_1s"] = (
                data_df["dam_nose_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_nose_movement_roll_mean_1ds"] = (
                data_df["dam_nose_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_nose_movement_roll_mean_2s"] = (
                data_df["dam_nose_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["eye_movement_roll_mean_1s"] = (
                data_df["eye_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["eye_movement_roll_mean_1ds"] = (
                data_df["eye_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["eye_movement_roll_mean_2s"] = (
                data_df["eye_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["ear_movement_roll_mean_1s"] = (
                data_df["ear_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["ear_movement_roll_mean_1ds"] = (
                data_df["ear_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["ear_movement_roll_mean_2s"] = (
                data_df["ear_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["top_head_dam_movement_roll_mean_1s"] = (
                data_df["top_head_dam_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["top_head_dam_movement_roll_mean_1ds"] = (
                data_df["top_head_dam_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["top_head_dam_movement_roll_mean_2s"] = (
                data_df["top_head_dam_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            # data_df['ventrum_side_movement_roll_mean_1s'] = data_df['ventrum_side_movement'].rolling(roll_windows[0], min_periods=0, center=True).mean()
            # data_df['ventrum_side_movement_roll_mean_1ds'] = data_df['ventrum_side_movement'].rolling(roll_windows[2], min_periods=0, center=True).mean()
            # data_df['ventrum_side_movement_roll_mean_2s'] = data_df['ventrum_side_movement'].rolling(roll_windows[4], min_periods=0, center=True).mean()

            # data_df['leg_front_movement_roll_mean_1s'] = data_df['leg_front_movement'].rolling(roll_windows[0], min_periods=0, center=True).mean()
            # data_df['leg_front_movement_roll_mean_1ds'] = data_df['leg_front_movement'].rolling(roll_windows[2], min_periods=0, center=True).mean()
            # data_df['leg_front_movement_roll_mean_2s'] = data_df['leg_front_movement'].rolling(roll_windows[4], min_periods=0, center=True).mean()

            # data_df['leg_behind_movement_roll_mean_1s'] = data_df['leg_behind_movement'].rolling(roll_windows[0], min_periods=0, center=True).mean()
            # data_df['leg_behind_movement_roll_mean_1ds'] = data_df['leg_behind_movement'].rolling(roll_windows[2], min_periods=0, center=True).mean()
            # data_df['leg_behind_movement_roll_mean_2s'] = data_df['leg_behind_movement'].rolling(roll_windows[4], min_periods=0, center=True).mean()

            data_df["wrist_movement_roll_mean_1s"] = (
                data_df["wrist_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["wrist_movement_roll_mean_1ds"] = (
                data_df["wrist_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["wrist_movement_roll_mean_2s"] = (
                data_df["wrist_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["armpit_movement_roll_mean_1s"] = (
                data_df["armpit_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["armpit_movement_roll_mean_1ds"] = (
                data_df["armpit_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["armpit_movement_roll_mean_2s"] = (
                data_df["armpit_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["dam_convex_hull_roll_mean_1s"] = (
                data_df["dam_convex_hull"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_convex_hull_roll_mean_1ds"] = (
                data_df["dam_convex_hull"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["dam_convex_hull_roll_mean_2s"] = (
                data_df["dam_convex_hull"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["head_convex_hull_roll_mean_1s"] = (
                data_df["head_convex_hull"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["head_convex_hull_roll_mean_1ds"] = (
                data_df["head_convex_hull"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["head_convex_hull_roll_mean_2s"] = (
                data_df["head_convex_hull"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["high_p_pup_bp_roll_mean_1s"] = (
                data_df["high_p_pup_bp"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["high_p_pup_bp_roll_mean_1ds"] = (
                data_df["high_p_pup_bp"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["high_p_pup_bp_roll_mean_2s"] = (
                data_df["high_p_pup_bp"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["high_p_dam_bp_roll_mean_1s"] = (
                data_df["high_p_dam_bp"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["high_p_dam_bp_roll_mean_1ds"] = (
                data_df["high_p_dam_bp"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["high_p_dam_bp_roll_mean_2s"] = (
                data_df["high_p_dam_bp"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            data_df["nose_back10_length_roll_mean_1s"] = (
                data_df["nose_back10_length"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .mean()
            )
            data_df["nose_back10_length_roll_mean_1ds"] = (
                data_df["nose_back10_length"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .mean()
            )
            data_df["nose_back10_length_roll_mean_2s"] = (
                data_df["nose_back10_length"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .mean()
            )

            print("Calculating rolling window sums ...")

            data_df["head_avg_movement_roll_sum_1s"] = (
                data_df["head_avg_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["head_avg_movement_roll_sum_1ds"] = (
                data_df["head_avg_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["head_avg_movement_roll_sum_2s"] = (
                data_df["head_avg_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["head_max_movement_roll_sum_1s"] = (
                data_df["head_max_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["head_max_movement_roll_sum_1ds"] = (
                data_df["head_max_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["head_max_movement_roll_sum_2s"] = (
                data_df["head_max_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["back_avg_movement_roll_sum_1s"] = (
                data_df["back_avg_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["back_avg_movement_roll_sum_1ds"] = (
                data_df["back_avg_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["back_avg_movement_roll_sum_2s"] = (
                data_df["back_avg_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["head_back_rel_roll_sum_1s"] = data_df[
                "head_avg_movement_roll_sum_1s"
            ] / (
                data_df["head_avg_movement_roll_sum_1s"]
                + data_df["back_avg_movement_roll_sum_1s"]
            )
            data_df["head_back_rel_roll_sum_1ds"] = data_df[
                "head_avg_movement_roll_sum_1ds"
            ] / (
                data_df["head_avg_movement_roll_sum_1ds"]
                + data_df["back_avg_movement_roll_sum_1ds"]
            )
            data_df["head_back_rel_roll_sum_2s"] = data_df[
                "head_avg_movement_roll_sum_2s"
            ] / (
                data_df["head_avg_movement_roll_sum_2s"]
                + data_df["back_avg_movement_roll_sum_2s"]
            )

            data_df["pups_convex_hull_roll_sum_1s"] = (
                data_df["pups_convex_hull"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["pups_convex_hull_roll_sum_1ds"] = (
                data_df["pups_convex_hull"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["pups_convex_hull_roll_sum_2s"] = (
                data_df["pups_convex_hull"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["dam_pup_distance_roll_sum_1s"] = (
                data_df["dam_pup_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_pup_distance_roll_sum_1ds"] = (
                data_df["dam_pup_distance"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_pup_distance_roll_sum_2s"] = (
                data_df["dam_pup_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["dam_pup30m_distance_roll_sum_1s"] = (
                data_df["dam_pup30m_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_pup30m_distance_roll_sum_1ds"] = (
                data_df["dam_pup30m_distance"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_pup30m_distance_roll_sum_2s"] = (
                data_df["dam_pup30m_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["dam_pup60m_distance_roll_sum_1s"] = (
                data_df["dam_pup60m_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_pup60m_distance_roll_sum_1ds"] = (
                data_df["dam_pup60m_distance"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_pup60m_distance_roll_sum_2s"] = (
                data_df["dam_pup60m_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["top_head_dam_movement_roll_sum_1s"] = (
                data_df["top_head_dam_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["top_head_dam_movement_roll_sum_1ds"] = (
                data_df["top_head_dam_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["top_head_dam_movement_roll_sum_2s"] = (
                data_df["top_head_dam_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["dam_back_angle_roll_sum_1s"] = (
                data_df["dam_back_angle"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_back_angle_roll_sum_1ds"] = (
                data_df["dam_back_angle"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_back_angle_roll_sum_2s"] = (
                data_df["dam_back_angle"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["dam_nose_movement_roll_sum_1s"] = (
                data_df["dam_nose_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_nose_movement_roll_sum_1ds"] = (
                data_df["dam_nose_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_nose_movement_roll_sum_2s"] = (
                data_df["dam_nose_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["eye_movement_roll_sum_1s"] = (
                data_df["eye_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["eye_movement_roll_sum_1ds"] = (
                data_df["eye_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["eye_movement_roll_sum_2s"] = (
                data_df["eye_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["ear_movement_roll_sum_1s"] = (
                data_df["ear_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["ear_movement_roll_sum_1ds"] = (
                data_df["ear_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["ear_movement_roll_sum_2s"] = (
                data_df["ear_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            # csv_df['ventrum_side_movement_roll_sum_1s'] = csv_df['ventrum_side_movement'].rolling(roll_windows[0], min_periods=0, center=True).sum()
            # csv_df['ventrum_side_movement_roll_sum_1ds'] = csv_df['ventrum_side_movement'].rolling(roll_windows[2], min_periods=0, center=True).sum()
            # csv_df['ventrum_side_movement_roll_sum_2s'] = csv_df['ventrum_side_movement'].rolling(roll_windows[4], min_periods=0, center=True).sum()

            # csv_df['leg_front_movement_roll_sum_1s'] = csv_df['leg_front_movement'].rolling(roll_windows[0], min_periods=0, center=True).sum()
            # csv_df['leg_front_movement_roll_sum_1ds'] = csv_df['leg_front_movement'].rolling(roll_windows[2], min_periods=0, center=True).sum()
            # csv_df['leg_front_movement_roll_sum_2s'] = csv_df['leg_front_movement'].rolling(roll_windows[4], min_periods=0, center=True).sum()

            # csv_df['leg_behind_movement_roll_sum_1s'] = csv_df['leg_behind_movement'].rolling(roll_windows[0], min_periods=0, center=True).sum()
            # csv_df['leg_behind_movement_roll_sum_1ds'] = csv_df['leg_behind_movement'].rolling(roll_windows[2], min_periods=0, center=True).sum()
            # csv_df['leg_behind_movement_roll_sum_2s'] = csv_df['leg_behind_movement'].rolling(roll_windows[4], min_periods=0, center=True).sum()

            data_df["wrist_movement_roll_sum_1s"] = (
                data_df["wrist_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["wrist_movement_roll_sum_1ds"] = (
                data_df["wrist_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["wrist_movement_roll_sum_2s"] = (
                data_df["wrist_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["armpit_movement_roll_sum_1s"] = (
                data_df["armpit_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["armpit_movement_roll_sum_1ds"] = (
                data_df["armpit_movement"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["armpit_movement_roll_sum_2s"] = (
                data_df["armpit_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["dam_convex_hull_roll_sum_1s"] = (
                data_df["dam_convex_hull"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_convex_hull_roll_sum_1ds"] = (
                data_df["dam_convex_hull"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["dam_convex_hull_roll_sum_2s"] = (
                data_df["dam_convex_hull"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            # csv_df['head_convex_hull_roll_sum_1s'] = csv_df['head_convex_hull'].rolling(roll_windows[0], min_periods=0, center=True).sum()
            # csv_df['head_convex_hull_roll_sum_1ds'] = csv_df['head_convex_hull'].rolling(roll_windows[2], min_periods=0, center=True).sum()
            # csv_df['head_convex_hull_roll_sum_2s'] = csv_df['head_convex_hull'].rolling(roll_windows[4], min_periods=0, center=True).sum()

            data_df["high_p_pup_bp_roll_sum_1s"] = (
                data_df["high_p_pup_bp"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["high_p_pup_bp_roll_sum_1ds"] = (
                data_df["high_p_pup_bp"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["high_p_pup_bp_roll_sum_2s"] = (
                data_df["high_p_pup_bp"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["high_p_dam_bp_roll_sum_1s"] = (
                data_df["high_p_dam_bp"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["high_p_dam_bp_roll_sum_1ds"] = (
                data_df["high_p_dam_bp"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["high_p_dam_bp_roll_sum_2s"] = (
                data_df["high_p_dam_bp"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            data_df["nose_back10_length_roll_sum_1s"] = (
                data_df["nose_back10_length"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .sum()
            )
            data_df["nose_back10_length_roll_sum_1ds"] = (
                data_df["nose_back10_length"]
                .rolling(roll_windows[2], min_periods=0, center=True)
                .sum()
            )
            data_df["nose_back10_length_roll_sum_2s"] = (
                data_df["nose_back10_length"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .sum()
            )

            print("Calculating rolling window standard deviations...")
            data_df["head_avg_movement_roll_std_1s"] = (
                data_df["head_avg_movement"]
                .rolling(roll_windows[0], min_periods=0)
                .std()
            )
            data_df["head_avg_movement_roll_std_2s"] = (
                data_df["head_avg_movement"]
                .rolling(roll_windows[4], min_periods=0)
                .std()
            )

            data_df["back_avg_movement_roll_std_1s"] = (
                data_df["back_avg_movement"]
                .rolling(roll_windows[0], min_periods=0)
                .std()
            )
            data_df["back_avg_movement_roll_std_2s"] = (
                data_df["back_avg_movement"]
                .rolling(roll_windows[4], min_periods=0)
                .std()
            )

            data_df["pups_convex_hull_roll_std_1s"] = (
                data_df["pups_convex_hull"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["pups_convex_hull_roll_std_2s"] = (
                data_df["pups_convex_hull"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["dam_pup_distance_roll_std_1s"] = (
                data_df["dam_pup_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["dam_pup_distance_roll_std_2s"] = (
                data_df["dam_pup_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["dam_pup30m_distance_roll_std_1s"] = (
                data_df["dam_pup30m_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["dam_pup30m_distance_roll_std_2s"] = (
                data_df["dam_pup30m_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["dam_pup60m_distance_roll_std_1s"] = (
                data_df["dam_pup60m_distance"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["dam_pup60m_distance_roll_std_2s"] = (
                data_df["dam_pup60m_distance"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["eye_movement_roll_std_1s"] = (
                data_df["eye_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["eye_movement_roll_std_2s"] = (
                data_df["eye_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["ear_movement_roll_std_1s"] = (
                data_df["ear_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["ear_movement_roll_std_2s"] = (
                data_df["ear_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["wrist_movement_roll_std_1s"] = (
                data_df["wrist_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["wrist_movement_roll_std_2s"] = (
                data_df["wrist_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["armpit_movement_roll_std_1s"] = (
                data_df["armpit_movement"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["armpit_movement_roll_std_2s"] = (
                data_df["armpit_movement"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["dam_convex_hull_roll_std_1s"] = (
                data_df["dam_convex_hull"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["dam_convex_hull_roll_std_2s"] = (
                data_df["dam_convex_hull"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["head_convex_hull_roll_std_1s"] = (
                data_df["head_convex_hull"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["head_convex_hull_roll_std_2s"] = (
                data_df["head_convex_hull"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["high_p_pup_bp_roll_std_1s"] = (
                data_df["high_p_pup_bp"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["high_p_pup_bp_roll_std_2s"] = (
                data_df["high_p_pup_bp"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["high_p_dam_bp_roll_std_1s"] = (
                data_df["high_p_dam_bp"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["high_p_dam_bp_roll_std_2s"] = (
                data_df["high_p_dam_bp"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["nose_back10_length_roll_std_1s"] = (
                data_df["nose_back10_length"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["nose_back10_length_roll_std_2s"] = (
                data_df["nose_back10_length"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["dam_head_angle_roll_std_1s"] = (
                data_df["dam_head_angle"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["dam_head_angle_roll_std_2s"] = (
                data_df["dam_head_angle"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["dam_head_angle2_roll_std_1s"] = (
                data_df["dam_head_angle2"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["dam_head_angle2_roll_std_2s"] = (
                data_df["dam_head_angle2"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            data_df["dam_back_top_angle_roll_std_1s"] = (
                data_df["dam_back_top_angle"]
                .rolling(roll_windows[0], min_periods=0, center=True)
                .std()
            )
            data_df["dam_back_top_angle_roll_std_2s"] = (
                data_df["dam_back_top_angle"]
                .rolling(roll_windows[4], min_periods=0, center=True)
                .std()
            )

            drop_cols = [
                "side_x",
                "arm_x",
                "head_centroid_x",
                "dam_centroid_x",
                "left_wrist_movement",
                "right_wrist_movement",
                "left_eye_movement",
                "right_eye_movement",
                "left_armpit_movement",
                "right_armpit_movement",
                "left_palm_movement",
                "right_palm_movement",
                "left_ear_movement",
                "right_ear_movement",
                "left_ankle_movement",
                "right_ankle_movement",
                "left_leg_behind_movement",
                "right_leg_behind_movement",
                "left_leg_front_movement",
                "right_leg_front_movement",
                "pups_centroid_x_roll_mean_30m",
                "pups_centroid_x_roll_mean_60m",
                "pups_centroid_y_roll_mean_30m",
                "pups_centroid_y_roll_mean_60m",
                "pups_centroid_mult_x",
                "pups_centroid_mult_y",
                "pups_centroid_x",
                "pups_centroid_y",
                "high_p_pup_bp_no_zero",
            ]

            data_df = data_df.drop(drop_cols, axis=1)

            print(f"Correcting coordinate units with {self.px_per_mm} pixels per mm...")
            correction_bps = ["dam_centroid", "head_centroid"]
            correction_columns = [bp + "_y" for bp in correction_bps]

            for col in correction_columns:
                data_df[col] = data_df[col] / self.px_per_mm

            data_df.replace([np.inf, -np.inf, np.nan], 0, inplace=True)
            data_df = data_df.fillna(0).apply(pd.to_numeric)

            print(f"Saving features for video {file_name}...")
            save_path = os.path.join(
                self.features_dir, file_name + "." + self.file_type
            )
            write_df(data_df, self.file_type, save_path)
            video_timer.stop_timer()
            print(
                f"Feature extraction complete for video {file_name} (elapsed time: {video_timer.elapsed_time_str}s)"
            )

        self.timer.stop_timer()
        stdout_success(
            f"Feature extraction complete for {str(len(self.files_found))} video(s). Results are saved inside the project_folder/csv/features_extracted directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


#
# extractor = AmberFeatureExtractor(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# extractor.run()



__author__ = "Simon Nilsson"

import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        FileSelect, SimbaButton)
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int)
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import InvalidInputError
from simba.utils.printing import stdout_success
from simba.utils.warnings import NoFileFoundWarning


class SetMachineModelParameters(PopUpMixin, ConfigReader):

    """
    :example:
    >>> _ = SetMachineModelParameters(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')

    """
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PopUpMixin.__init__(self, title="SET MODEL PARAMETERS", icon='equation_small')
        self.clf_table_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.SET_RUN_ML_PARAMETERS.value)
        Label(self.clf_table_frm,text="CLASSIFIER",font=Formats.FONT_HEADER.value).grid(row=0, column=0)
        Label(self.clf_table_frm, text="MODEL PATH (.SAV)", font=Formats.FONT_HEADER.value).grid(row=0, column=1, sticky=NW)
        Label(self.clf_table_frm, text="DISCRIMINATION THRESHOLD", font=Formats.FONT_HEADER.value).grid(row=0, column=2, sticky=NW)
        Label(self.clf_table_frm, text="MINIMUM BOUT LENGTH (MS)", font=Formats.FONT_HEADER.value).grid(row=0, column=3, sticky=NW)
        self.clf_data = {}
        for clf_cnt, clf_name in enumerate(self.clf_names):
            self.clf_data[clf_name] = {}
            Label(self.clf_table_frm, text=clf_name, font=Formats.FONT_HEADER.value).grid(row=clf_cnt + 1, column=0, sticky=NW)
            self.clf_data[clf_name]["path"] = FileSelect(self.clf_table_frm, title="Select model (.sav) file", initialdir=self.project_path, file_types=[("SimBA Classifier", "*.sav")])
            self.clf_data[clf_name]["threshold"] = Entry_Box(self.clf_table_frm, "", "0")
            self.clf_data[clf_name]["min_bout"] = Entry_Box(self.clf_table_frm, "", "0")
            self.clf_data[clf_name]["path"].grid(row=clf_cnt + 1, column=1, sticky=NW)
            self.clf_data[clf_name]["threshold"].grid(row=clf_cnt + 1, column=2, sticky=NW)
            self.clf_data[clf_name]["min_bout"].grid(row=clf_cnt + 1, column=3, sticky=NW)

        set_btn = SimbaButton(parent=self.main_frm, txt="SET MODEL(S)", img='tick', txt_clr='red', font=Formats.FONT_REGULAR.value, cmd=self.set)
        self.clf_table_frm.grid(row=0, sticky=W, pady=5, padx=5)
        set_btn.grid(row=1, pady=10)
        self.main_frm.mainloop()

    def set(self):
        filtered_clf_data = {}
        for model_name, model_settings in self.clf_data.items():
            print(model_settings)
            valid_path = check_file_exist_and_readable(model_settings["path"].file_path, raise_error=False)
            valid_threshold, _ = check_float(name=f"Classifier {model_name} threshold", value=model_settings["threshold"].entry_get, max_value=1.0, min_value=0.0, raise_error=False)
            valid_min_bout, _ = check_int(name=f"Classifier {model_name} minimum bout", value=model_settings["min_bout"].entry_get, min_value=0, raise_error=False)
            if not valid_path:
                NoFileFoundWarning(msg=f'SKIPPING CLASSIFIER {model_name}: The path "{model_settings["path"].file_path}" is not a valid file path for classifier {model_name}', source=self.__class__.__name__)
                filtered_clf_data[model_name] = {'path': None, 'threshold': None, 'min_bout': None}
            elif not valid_threshold:
                NoFileFoundWarning(msg=f'SKIPPING CLASSIFIER {model_name}: Classifier {model_name} threshold {model_settings["threshold"].entry_get} is not a valid threshold between 0.0 - 1.0.', source=self.__class__.__name__)
                filtered_clf_data[model_name] = {'path': None, 'threshold': None, 'min_bout': None}
            elif not valid_min_bout:
                NoFileFoundWarning(msg=f'SKIPPING CLASSIFIER {model_name}: Classifier {model_name} minimum bout length {model_settings["min_bout"].entry_get} is not a valid integer number', source=self.__class__.__name__)
                filtered_clf_data[model_name] = {'path': None, 'threshold': None, 'min_bout': None}
            else:
                filtered_clf_data[model_name] = {'path': model_settings["path"].file_path, 'threshold': model_settings["threshold"].entry_get, 'min_bout': model_settings["min_bout"].entry_get}
        if len(list(filtered_clf_data.keys())) == 0:
            raise InvalidInputError(msg=f'None of the {len(self.clf_names)} classifier(s) have valid paths, minimum bout lengths, and/or thresholds', source=self.__class__.__name__)

        for cnt, (model_name, model_settings) in enumerate(filtered_clf_data.items()):
            self.config.set("SML settings", f"model_path_{cnt + 1}", str(model_settings["path"]))
            self.config.set("threshold_settings", f"threshold_{cnt + 1}", str(model_settings["threshold"]))
            self.config.set("Minimum_bout_lengths", f"min_bout_{cnt + 1}", str(model_settings["min_bout"]))

        with open(self.config_path, "w") as f:
            self.config.write(f)

        stdout_success(msg="Model paths/settings saved in project_config.ini", source=self.__class__.__name__)


#_ = SetMachineModelParameters(config_path=r"C:\troubleshooting\Top_down_old\project_folder\project_config.ini")



__author__ = "Simon Nilsson"

import functools
import itertools
import multiprocessing
import os
from typing import List, Optional, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_file_exist_and_readable, check_int,
                                check_valid_boolean, check_valid_dataframe,
                                check_valid_lst)
from simba.utils.data import (create_color_palettes, detect_bouts,
                              slice_roi_dict_from_attribute)
from simba.utils.enums import Defaults, Formats, Keys, TextOptions
from simba.utils.errors import NoROIDataError, NoSpecifiedOutputError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    create_directory, find_core_cnt,
                                    get_fn_ext, get_video_meta_data, read_df,
                                    read_frm_of_video, remove_a_folder)


def _plot_cue_light_data(frm_idxs: list,
                         video_setting: bool,
                         frame_setting: bool,
                         show_pose: bool,
                         data_df: pd.DataFrame,
                         bp_names: list,
                         font_size: int,
                         x_shift: int,
                         y_shift: int,
                         frames_save_dir: str,
                         video_save_dir: str,
                         circle_size: int,
                         roi_dict: dict,
                         video_path: str,
                         verbose: bool):

    batch_id, frame_rng = frm_idxs[0], frm_idxs[1]
    start_frm, end_frm, current_frm = frame_rng[0], frame_rng[-1], frame_rng[0]
    video_writer = None
    video_meta_data = get_video_meta_data(video_path=video_path)

    clrs = create_color_palettes(no_animals=1, map_size=len(bp_names)+1, cmaps=['Set3'])[0]
    if video_setting:
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        video_save_path = os.path.join(video_save_dir, f"{batch_id}.mp4")
        video_writer = cv2.VideoWriter(video_save_path, fourcc, video_meta_data['fps'], (int(video_meta_data['width']*2), video_meta_data['height']))


    while current_frm <= end_frm:
        img = read_frm_of_video(video_path, frame_index=current_frm)
        img = cv2.copyMakeBorder(img, 0, 0, 0, int(video_meta_data["width"]), borderType=cv2.BORDER_CONSTANT, value=[0, 0, 0])
        if show_pose:
            for bp_cnt, bp_name in enumerate(bp_names):
                col_names = [f'{bp_name}_x', f'{bp_name}_y']
                bp_data = data_df.loc[current_frm, col_names].values.astype(np.int32)
                img = cv2.circle(img, tuple(bp_data), circle_size, clrs[bp_cnt], -1)
        img = PlottingMixin().roi_dict_onto_img(img=img, roi_dict=roi_dict)

        y_shift_counts = 1
        for cue_light_type, cue_light_type_data in roi_dict.items():
            for _, cue_light_data in cue_light_type_data.iterrows():
                color, name = cue_light_data['Color BGR'], cue_light_data['Name']
                light_status = 'ON' if data_df.loc[current_frm, name] == 1 else 'OFF'
                light_color = (0, 255, 255) if light_status == 'ON' else color
                que_light_bouts = detect_bouts(data_df=data_df.loc[0:current_frm], target_lst=[name], fps=video_meta_data['fps'])
                que_light_bouts_cnt = len(que_light_bouts)
                que_light_bouts_duration = round(que_light_bouts['Bout_time'].sum(), 2)
                off_duration = round((((current_frm + 1)/ video_meta_data['fps'])) - que_light_bouts_duration, 2)
                img = PlottingMixin().put_text(img=img, text=f"{name} STATUS:", pos=(int(video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), int(y_shift*y_shift_counts)), font_size=font_size, font_thickness=2, text_color_bg=(0, 0, 0), text_color=color)
                img = PlottingMixin().put_text(img=img, text=light_status, pos=(int(video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value + x_shift), int(y_shift*y_shift_counts)), font_size=font_size, font_thickness=2, text_color_bg=(0, 0, 0), text_color=light_color)
                y_shift_counts += 1
                img = PlottingMixin().put_text(img=img, text=f"{name} ONSET COUNTS:", pos=(int(video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), int(y_shift*y_shift_counts)), font_size=font_size, font_thickness=2, text_color_bg=(0, 0, 0), text_color=color)
                img = PlottingMixin().put_text(img=img, text=str(que_light_bouts_cnt), pos=(int(video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value + x_shift), int(y_shift*y_shift_counts)), font_size=font_size, font_thickness=2, text_color_bg=(0, 0, 0), text_color=color)
                y_shift_counts += 1
                img = PlottingMixin().put_text(img=img, text=f"{name} TIME ON (S):", pos=(int(video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), int(y_shift*y_shift_counts)), font_size=font_size, font_thickness=2, text_color_bg=(0, 0, 0), text_color=color)
                img = PlottingMixin().put_text(img=img, text=str(que_light_bouts_duration), pos=(int(video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value + x_shift), int(y_shift*y_shift_counts)), font_size=font_size, font_thickness=2, text_color_bg=(0, 0, 0), text_color=color)
                y_shift_counts += 1
                img = PlottingMixin().put_text(img=img, text=f"{name} TIME OFF (S):", pos=(int(video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), int(y_shift*y_shift_counts)), font_size=font_size, font_thickness=2, text_color_bg=(0, 0, 0), text_color=color)
                img = PlottingMixin().put_text(img=img, text=str(off_duration), pos=(int(video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value + x_shift), int(y_shift*y_shift_counts)), font_size=font_size, font_thickness=2, text_color_bg=(0, 0, 0), text_color=color)
                y_shift_counts += 1

        if video_setting:
            video_writer.write(np.uint8(img))
        if frame_setting:
            frame_save_path = os.path.join(frames_save_dir,f"{current_frm}.png")
            cv2.imwrite(frame_save_path, current_frm)
        if verbose:
            print(f"Cue light frame complete: {current_frm} / {video_meta_data['frame_count']}. Video: {video_meta_data['video_name']} ")
        current_frm += 1
    if video_setting:
        video_writer.release()
    return batch_id

class CueLightVisualizer(ConfigReader):
    """

    Visualize SimBA computed cue-light ON and OFF states and the aggregate statistics of ON and OFF
    states.

    :param str config_path: path to SimBA project config file in Configparser format.
    :param List[str] cue_light_names: Names of cue lights, as defined in the SimBA ROI interface.
    :param str video_path: Path to video which user wants to create visualizations of cue light states and aggregate statistics for.
    :param bool frame_setting: If True, creates individual frames in png format. Defaults to False.
    :param bool video_setting: If True, creates compressed videos in mp4 format. Defaults to True.

    .. notes:
       `Cue light tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md>`__.

    .. video:: _static/img/CueLightVisualizer.webm
       :width: 800
       :autoplay:
       :loop:

    :examples:
    >>> cue_light_visualizer = CueLightVisualizer(config_path='SimBAConfig', cue_light_names=['Cue_light'], video_path='VideoPath', video_setting=True, frame_setting=False)
    >>> cue_light_visualizer.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 cue_light_names: List[str],
                 video_path: Union[str, os.PathLike],
                 data_path: Union[str, os.PathLike],
                 frame_setting: bool = False,
                 video_setting: bool = True,
                 core_cnt: int = -1,
                 show_pose: bool = True,
                 verbose: bool = True):

        ConfigReader.__init__(self, config_path=config_path)
        check_valid_boolean(value=[frame_setting], source=f'{self.__class__.__name__} frame_setting', raise_error=True)
        check_valid_boolean(value=[video_setting], source=f'{self.__class__.__name__} video_setting', raise_error=True)
        check_valid_boolean(value=[show_pose], source=f'{self.__class__.__name__} show_pose', raise_error=True)
        check_valid_boolean(value=[verbose], source=f'{self.__class__.__name__} verbose', raise_error=True)
        check_valid_lst(data=cue_light_names, source=self.__class__.__name__, valid_dtypes=(str,), min_len=1, raise_error=True)
        check_file_exist_and_readable(file_path=video_path)
        check_file_exist_and_readable(file_path=data_path)
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1, unaccepted_vals=[0])
        self.video_meta_data = get_video_meta_data(video_path)
        _, self.video_name, _ = get_fn_ext(filepath=data_path)
        if (not frame_setting) and (not video_setting):
            raise NoSpecifiedOutputError(msg="SIMBA ERROR: Please choose to select either videos, frames, or both frames and videos.")
        self.cue_light_names, self.video_path, self.data_path = cue_light_names, video_path, data_path
        self.data_df = read_df(self.data_path, self.file_type)
        self.core_cnt = find_core_cnt()[0] if core_cnt == -1 or core_cnt > find_core_cnt()[0] else core_cnt
        self.font_size, self.x_shift, self.y_shift = PlottingMixin().get_optimal_font_scales(text='ONE LONG ARSE STRING FOR YA', accepted_px_height=int(self.video_meta_data['height']/2), accepted_px_width=int(self.video_meta_data['width']/2))
        self.circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(self.video_meta_data['width'], self.video_meta_data['height']), circle_frame_ratio=60)
        self.read_roi_data()
        self.video_setting, self.frame_setting, self.data_path, self.show_pose, self.verbose = video_setting, frame_setting, data_path, show_pose, verbose
        self.video_save_dir = os.path.join(self.frames_output_dir, 'cue_lights')
        self.frames_save_dir = os.path.join(self.frames_output_dir, 'cue_lights')
        self.video_roi_dict, roi_names, video_roi_cnt = slice_roi_dict_from_attribute(data=self.roi_dict, shape_names=self.cue_light_names, video_names=[self.video_name])
        missing_rois = [x for x in self.cue_light_names if x not in roi_names]
        if len(missing_rois) > 0:
            raise NoROIDataError(msg=f'The video {self.video_name} does not have cue light ROI(s) named {missing_rois}.', source=self.__class__.__name__)
        if show_pose:
            check_valid_dataframe(df=self.data_df, source=f'{self.__class__.__name__} {data_path}', valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.bp_col_names)


    def run(self):
        print(f"Creating video for {len(self.cue_light_names)} cue light(s) in in {self.video_name}...")
        frames_dir, video_temp_dir = None, None
        if self.frame_setting:
            frames_dir = os.path.join(self.frames_save_dir, self.video_name)
            create_directory(paths=frames_dir, overwrite=True)
        if self.video_setting:
            self.save_video_path = os.path.join(self.video_save_dir, f"{self.video_name}.mp4")
            video_temp_dir = os.path.join(self.video_save_dir, 'temp')
            create_directory(paths=video_temp_dir, overwrite=True)
        self.frm_lst = list(range(0, self.video_meta_data["frame_count"], 1))
        self.frame_chunks = np.array_split(self.frm_lst, self.core_cnt)
        self.frame_chunks = [(x, j) for x, j in enumerate(self.frame_chunks)]
        with multiprocessing.Pool(self.core_cnt, maxtasksperchild=Defaults.MAXIMUM_MAX_TASK_PER_CHILD.value) as pool:
            constants = functools.partial(_plot_cue_light_data,
                                          frame_setting=self.frame_setting,
                                          video_setting=self.video_setting,
                                          show_pose=self.show_pose,
                                          data_df=self.data_df,
                                          frames_save_dir=frames_dir,
                                          video_save_dir=video_temp_dir,
                                          circle_size=self.circle_size,
                                          font_size=self.font_size,
                                          x_shift=self.x_shift,
                                          y_shift=self.y_shift,
                                          roi_dict=self.video_roi_dict,
                                          video_path=self.video_path,
                                          bp_names=self.body_parts_lst,
                                          verbose=self.verbose)
            for cnt, result in enumerate(pool.imap(constants, self.frame_chunks, chunksize=self.multiprocess_chunksize)):
                if self.verbose:
                    print(f'Batch {int(result+1/self.core_cnt)} complete...')
            pool.terminate()
            pool.join()
        self.timer.stop_timer()
        if self.video_setting:
            print(f"Joining {self.video_name} multiprocessed video...")
            concatenate_videos_in_folder(in_folder=video_temp_dir, save_path=self.save_video_path)
            stdout_success(msg=f"Cue light video visualization for video {self.video_name} saved at {self.save_video_path}", elapsed_time=self.timer.elapsed_time_str)
        if self.frame_setting:
            stdout_success(msg=f"Cue light frame visualization for video {self.video_name} saved at {frames_dir}", elapsed_time=self.timer.elapsed_time_str)

# if __name__ == "__main__":
#     test = CueLightVisualizer(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini",
#                               cue_light_names=['cl'],
#                               video_path=r"C:\troubleshooting\cue_light\t1\project_folder\videos\2025-05-21 16-10-06_cropped.mp4",
#                               data_path=r"C:\troubleshooting\cue_light\t1\project_folder\csv\cue_lights\2025-05-21 16-10-06_cropped.csv",
#                               video_setting=True,
#                               frame_setting=False,
#                               core_cnt=23)
#
#     test.run()


__author__ = "Simon Nilsson"

import os
from typing import List, Optional, Union

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_supplement_mixin import \
    FeatureExtractionSupplemental
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log, check_if_dir_exists,
    check_int, check_str, check_valid_dataframe, check_valid_lst)
from simba.utils.data import detect_bouts
from simba.utils.enums import Formats
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df)


class CueLightMovementAnalyzer(ConfigReader):
    """
    Compute aggregate statistics of animal movement in relation to the cue light
    ON and OFF states.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter int pre_window: Time period (in millisecond) before the onset of each cue light to compute aggregate classification statistics within.
    :parameter int post_window: Time period (in millisecond) after the offset of each cue light to compute aggregate classification statistics within.
    :parameter List[str] cue_light_names: Names of cue lights, as defined in the SimBA ROI interface.
    :parameter float threshold: The body-part post-estimation probability threshold. SimBA omits movement calculations for frames where the body-part probability threshold is lower than the user-specified threshold.
    :parameter bool roi_setting: If True, SimBA calculates movement statistics within non-light cue ROIs.

    .. note:
       `Cue light tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md>`__.


    :examples:
    >>> cue_light_movement_analyzer = CueLightMovementAnalyzer(config_path='MyProjectConfig', cue_light_names=['Cue_light'], pre_window=1000, post_window=1000, threshold=0.0, roi_setting=True)
    >>> cue_light_movement_analyzer.calculate_whole_session_movement()
    >>> cue_light_movement_analyzer.organize_results()
    >>> cue_light_movement_analyzer.save_results()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 cue_light_names: List[str],
                 bp_name: str,
                 data_dir: Optional[Union[str, os.PathLike]] = None,
                 pre_window: int = 0,
                 post_window: int = 0):

        ConfigReader.__init__(self, config_path=config_path)
        check_valid_lst(data=cue_light_names, source=f'{self.__class__.__name__} cue_light_names', valid_dtypes=(str,), min_len=1, raise_error=True)
        check_int(name=f'{self.__class__.__name__} pre_window', value=pre_window, min_value=0)
        check_int(name=f'{self.__class__.__name__} post_window', value=post_window, min_value=0)
        if data_dir is None:
            self.data_dir = self.cue_lights_data_dir
        else:
            check_if_dir_exists(in_dir=data_dir)
            self.data_dir = data_dir
        self.data_paths = find_files_of_filetypes_in_directory(directory=self.data_dir, extensions=[f'.{self.file_type}'], raise_error=True, as_dict=True)
        self.data_file_count = len(list(self.data_paths.keys()))
        check_str(name=f'{self.__class__.__name__} bp_name', value=bp_name, options=self.body_parts_lst)
        self.bp_cols = [f'{bp_name}_x', f'{bp_name}_y']
        self.cue_light_names, self.pre_window, self.post_window, self.bp_name = cue_light_names, pre_window, post_window, bp_name
        self.save_path = os.path.join(self.logs_path, f"Cue_lights_movement_statistics_{self.datetime}.csv")

    def run(self):
        print(f"Analyzing cue-light movement data for {self.data_file_count} file(s)...")
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=list(self.data_paths.values()))
        self.results = pd.DataFrame(columns=['VIDEO', 'BODY PART', 'CUE LIGHT', 'CUE LIGHT BOUT START TIME', 'CUE LIGHT BOUT END TIME', 'CUE LIGHT BOUT START FRAME', 'CUE LIGHT BOUT END FRAME', 'CUE LIGHT BOUT MOVEMENT (CM)', 'CUE LIGHT BOUT VELOCITY (CM/S)', f'PRE CUE LIGHT BOUT ({self.pre_window}s) MOVEMENT (CM)', f'PRE CUE LIGHT BOUT ({self.pre_window}s) VELOCITY (CM/S)', f'POST CUE LIGHT BOUT ({self.post_window}s) MOVEMENT (CM)', f'POST CUE LIGHT BOUT ({self.post_window}s) VELOCITY (CM/S)'])
        for file_cnt, (video_name, data_path) in enumerate(self.data_paths.items()):
            video_timer = SimbaTimer(start=True)
            self.data_df = read_df(data_path, self.file_type).reset_index(drop=True)
            check_valid_dataframe(df=self.data_df, source=data_path, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.bp_cols + self.cue_light_names)
            _, px_per_mm, fps = self.read_video_info(video_name=video_name)
            self.prior_window_frames_cnt = int(self.pre_window * fps)
            self.post_window_frames_cnt = int(self.post_window * fps)
            cue_light_bouts = detect_bouts(data_df=self.data_df, target_lst=self.cue_light_names, fps=fps).reset_index(drop=True)
            for bout_cnt, bout in cue_light_bouts.iterrows():
                cue_frm_range = list(range(bout['Start_frame'], bout['End_frame']+1))
                pre_window_frms = list(range(max(0, bout['Start_frame']-self.prior_window_frames_cnt), bout['Start_frame']))
                post_window_frms = list(range(bout['End_frame']+1, min((bout['End_frame'] + self.post_window_frames_cnt), len(self.data_df))))
                cue_frm_range_arr = self.data_df.loc[cue_frm_range][self.bp_cols].values
                pre_window_frms_arr = self.data_df.loc[pre_window_frms][self.bp_cols].values
                post_window_frms_arr = self.data_df.loc[post_window_frms][self.bp_cols].values
                cue_bout_movement, cue_bout_velocity = FeatureExtractionSupplemental().distance_and_velocity(x=cue_frm_range_arr, fps=fps, pixels_per_mm=px_per_mm, centimeters=True)
                pre_window_movement, pre_window_velocity = FeatureExtractionSupplemental().distance_and_velocity(x=pre_window_frms_arr, fps=fps, pixels_per_mm=px_per_mm, centimeters=True)
                post_window_movement, post_window_velocity = FeatureExtractionSupplemental().distance_and_velocity(x=post_window_frms_arr, fps=fps, pixels_per_mm=px_per_mm, centimeters=True)
                self.results.loc[len(self.results)] = [video_name, self.bp_name, bout['Event'], bout['Start_time'], bout['End Time'], bout['Start_frame'], bout['End_frame'], cue_bout_movement, cue_bout_velocity, pre_window_movement, pre_window_velocity, post_window_movement, post_window_velocity]
            video_timer.stop_timer()
            print(f'Cue light movement statistics for video {video_name} complete... ({file_cnt+1}/{self.data_file_count}, elapsed time: {video_timer.elapsed_time_str}s)')
    def save(self):
        self.results = self.results.sort_values(by=['VIDEO', 'CUE LIGHT', 'CUE LIGHT BOUT START TIME'], ascending=True)
        if self.post_window == 0:
            self.results = self.results.drop([f'POST CUE LIGHT BOUT ({self.post_window}s) MOVEMENT (CM)', f'POST CUE LIGHT BOUT ({self.post_window}s) VELOCITY (CM/S)'], axis=1)
        if self.pre_window == 0:
            self.results = self.results.drop([f'PRE CUE LIGHT BOUT ({self.pre_window}s) MOVEMENT (CM)', f'PRE CUE LIGHT BOUT ({self.pre_window}s) VELOCITY (CM/S)'], axis=1)
        self.results.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(msg=f'Cue light movement data saved at {self.save_path}', elapsed_time=self.timer.elapsed_time_str)




# test = CueLightMovementAnalyzer(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini",
#                                 pre_window=0,
#                                 post_window=0,
#                                 cue_light_names=['cl'],
#                                 data_dir=r"C:\troubleshooting\cue_light\t1\project_folder\csv\cue_lights",
#                                 bp_name='Nose')
#
# test.run()
# test.save()

# test = CueLightMovementAnalyzer(config_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
#                                 pre_window=100, post_window=100, cue_light_names=['Rectangl_1'], threshold=0.0, roi_setting=True)
#
#
# def __init__(self,
#              config_path: str,
#              pre_window: int,
#              post_window: int,
#              cue_light_names: list,
#              threshold: float,
#              roi_setting: bool):



import os
from typing import Optional, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.statistics_mixin import Statistics
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.errors import CountError
from simba.utils.read_write import read_pickle
from simba.utils.warnings import CountWarning

LOF = "local outlier factor"
EE = "elliptic envelope"


class OutlierDetector(ConfigReader):

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_path: Union[str, os.PathLike],
        algorithm: Literal[LOF, EE],
        cluster_sliced: Optional[bool] = True,
    ):

        ConfigReader.__init__(
            self, config_path=config_path, read_video_info=False, create_logger=False
        )
        data = read_pickle(data_path=data_path, verbose=True)
        self.x = data[Unsupervised.DR_MODEL.value]["MODEL"].embedding_
        idx = np.array(
            list(map(np.array, np.array(data["DATA"]["BOUTS_FEATURES"].index.values)))
        )
        self.x = np.hstack((idx, self.x))
        if cluster_sliced:
            self.y = data[Clustering.CLUSTER_MODEL.value][
                Unsupervised.MODEL.value
            ].labels_
            self.unique_labels = [x for x in list(np.unique(self.y)) if x != -1]
            if len(self.unique_labels) <= 1:
                CountWarning(
                    msg="Too few clusters for performing cluster_sliced, reverting to False",
                    source=self.__class__.__name__,
                )
                cluster_sliced = False
        else:
            self.y = None
        self.cluster_sliced, self.algorithm = cluster_sliced, algorithm

    def run(self):
        if self.cluster_sliced:
            self.results = []
            for i in self.unique_labels:
                c = self.x[np.argwhere(self.y == i)].reshape(-1, 5)
                c_id, c = c[:, 0:3], c[:, 3:5].astype(np.float32)
        #         if self.algorithm is LOF:
        #             s = Statistics.local_outlier_factor(data=c, k=0.2, normalize=True)
        #         elif self.algorithm is EE:
        #             s = Statistics.elliptic_envelope(data=c, normalize=True)
        #         self.results.append(np.hstack([c, np.full((c.shape[0], 1), i).reshape(-1, 1), s.reshape(-1, 1)]))
        #     self.results = np.concatenate(self.results, axis=0)
        #     unclustered = self.x[np.argwhere(self.y == -1)].reshape(-1, 2)
        #     if unclustered.shape[0] > 0:
        #         unclustered = np.hstack((unclustered, np.full((unclustered.shape[0], 1), -1), np.full((unclustered.shape[0], 1), np.max(self.results[:, 3]))))
        #         self.results = np.vstack((self.results, unclustered))
        #
        # else:
        #     if self.algorithm is LOF:
        #         s = Statistics.local_outlier_factor(data=self.x, k=0.2, normalize=True)
        #     elif self.algorithm is EE:
        #         s = Statistics.elliptic_envelope(data=self.x, normalize=True)
        #     self.results = np.vstack([self.x, s.reshape(-1, 1)])
        #
        #
        # print(self.results)

        # img = PlottingMixin.continuous_scatter(data=self.results, columns=('X', 'Y', 'LOF'), size=20, palette='seismic')

        # results = pd.DataFrame(np.vstack((results, unclustered)), columns=('X', 'Y', 'CLUSTER', 'LOF'))

        # img = PlottingMixin.continuous_scatter(data=self.results, columns=('X', 'Y', 'LOF'), size=20, palette='seismic')

        #
        #
        #
        #
        #
        # results = []
        # for i in [x for x in unique_clusters if x != -1]:
        #     c = x[np.argwhere(y == i)].reshape(-1, 2)
        #     lof = Statistics.local_outlier_factor(data=c, k=100)
        #     results.append(np.hstack([c, np.full((c.shape[0], 1), i).reshape(-1, 1), lof.reshape(-1, 1)]))
        #
        # results = np.concatenate(results, axis=0)
        # unclustered = x[np.argwhere(y == -1)].reshape(-1, 2)
        # unclustered = np.hstack((unclustered, np.full((unclustered.shape[0], 1), -1), np.full((unclustered.shape[0], 1), np.max(results[:, 3]))))
        # results = pd.DataFrame(np.vstack((results, unclustered)), columns=('X', 'Y', 'CLUSTER', 'LOF'))
        #
        # #results['LOF'] = (results['LOF'] - results['LOF'].min()) / (y - x) * (b - a) + a
        #
        # results = (results - results.mean()) / results.std()
        # results=(results-results.min())/(results.max()-results.min())
        #
        #
        # img = PlottingMixin.continuous_scatter(data=results, columns=('X', 'Y', 'LOF'), size=20, palette='seismic')
        #
        #


# data_path = '/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/clusters/beautiful_beaver.pickle'
data_path = "/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/small_clusters/adoring_hoover.pickle"
x = OutlierDetector(
    data_path=data_path,
    config_path="/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini",
    algorithm=LOF,
    cluster_sliced=True,
)
x.run()



__author__ = "Simon Nilsson"

import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.pose_processors.pose_reset import PoseResetter
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, FileSelect, SimbaButton,
                                        TwoOptionQuestionPopUp)
from simba.utils.checks import check_str
from simba.utils.enums import ConfigKey, Keys, Links
from simba.utils.errors import DuplicationError
from simba.utils.printing import stdout_success, stdout_trash
from simba.utils.read_write import tabulate_clf_info


class AddClfPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, config_path=config_path, title="ADD CLASSIFIER")
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.clf_eb = Entry_Box(self.main_frm, "CLASSIFIER NAME:", "15")
        add_btn = SimbaButton(parent=self.main_frm, txt="ADD CLASSIFIER", cmd=self.run, img='rocket')
        self.clf_eb.grid(row=0, column=0, sticky=NW)
        add_btn.grid(row=1, column=0, sticky=NW)

    def run(self):
        clf_name = self.clf_eb.entry_get.strip()
        check_str(name="CLASSIFIER NAME", value=clf_name)
        if clf_name in self.clf_names:
            raise DuplicationError(msg=f'The classifier name {clf_name} already exist in the SimBA project.', source=self.__class__.__name__)
        self.config.set( ConfigKey.SML_SETTINGS.value, ConfigKey.TARGET_CNT.value, str(self.clf_cnt + 1))
        self.config.set(ConfigKey.SML_SETTINGS.value, f"model_path_{str(self.clf_cnt + 1)}", "")
        self.config.set(ConfigKey.SML_SETTINGS.value, f"target_name_{str(self.clf_cnt + 1)}", clf_name)
        self.config.set(ConfigKey.THRESHOLD_SETTINGS.value, f"threshold_{str(self.clf_cnt + 1)}", "None")
        self.config.set(ConfigKey.MIN_BOUT_LENGTH.value, f"min_bout_{str(self.clf_cnt + 1)}", "None")
        with open(self.config_path, "w") as f:
            self.config.write(f)
        stdout_success(msg=f"{clf_name} classifier added to SimBA project", source=self.__class__.__name__)


class RemoveAClassifierPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="Warning: Remove classifier(s) settings")
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.remove_clf_frm = CreateLabelFrameWithIcon( parent=self.main_frm, header="SELECT A CLASSIFIER TO REMOVE", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.REMOVE_CLF.value)
        self.clf_dropdown = DropDownMenu(self.remove_clf_frm, "Classifier", self.clf_names, "12")
        self.clf_dropdown.setChoices(self.clf_names[0])

        run_btn = SimbaButton(parent=self.main_frm, txt="REMOVE CLASSIFIER", cmd=self.run, img='trash')
        self.remove_clf_frm.grid(row=0, sticky=W)
        self.clf_dropdown.grid(row=0, sticky=W)
        run_btn.grid(row=1, pady=10)

    def run(self):
        for i in range(len(self.clf_names)):
            self.config.remove_option("SML settings", f"model_path_{i+1}")
            self.config.remove_option("SML settings", f"target_name_{i+1}")
            self.config.remove_option("threshold_settings", f"threshold_{i+1}")
            self.config.remove_option("Minimum_bout_lengths", f"min_bout_{i+1}")
        self.clf_names.remove(self.clf_dropdown.getChoices())
        self.config.set("SML settings", "no_targets", str(len(self.clf_names)))

        for clf_cnt, clf_name in enumerate(self.clf_names):
            self.config.set("SML settings", f"model_path_{clf_cnt + 1}", "")
            self.config.set("SML settings", f"target_name_{clf_cnt + 1}", clf_name)
            self.config.set("threshold_settings", f"threshold_{clf_cnt + 1}", "None")
            self.config.set("Minimum_bout_lengths", f"min_bout_{clf_cnt + 1}", "None")

        with open(self.config_path, "w") as f:
            self.config.write(f)

        stdout_trash(msg=f"{self.clf_dropdown.getChoices()} classifier removed from SimBA project.", source=self.__class__.__name__)


# _ = RemoveAClassifierPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini')


class PrintModelInfoPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="PRINT MACHINE MODEL INFO", size=(250, 250))
        model_info_frame = LabelFrame(
            self.main_frm, text="PRINT MODEL INFORMATION", padx=5, pady=5, font="bold"
        )
        model_path_selector = FileSelect(
            model_info_frame, "Model path", title="Select a video file"
        )
        btn_print_info = Button(
            model_info_frame,
            text="PRINT MODEL INFO",
            command=lambda: tabulate_clf_info(clf_path=model_path_selector.file_path),
        )
        model_info_frame.grid(row=0, sticky=W)
        model_path_selector.grid(row=0, sticky=W, pady=5)
        btn_print_info.grid(row=1, sticky=W)


class PoseResetterPopUp:
    def __init__(self):
        # PopUpMixin.__init__(self, title="WARNING!", size=(300, 100))
        # popupframe = LabelFrame(self.main_frm)
        # label = Label(popupframe, text="Do you want to remove user-defined pose-configurations?")
        # label.grid(row=0, columnspan=2)
        # B1 = Button(popupframe,text="YES",fg="blue",command=lambda: PoseResetter(master=self.main_frm))
        # B2 = Button(popupframe, text="NO", fg="red", command=self.main_frm.destroy)
        question = TwoOptionQuestionPopUp(
            title="WARNING!",
            question="Do you want to remove user-defined pose-configurations?",
            option_one="YES",
            option_two="NO",
        )
        if question.selected_option == "YES":
            _ = PoseResetter(master=None)
        else:
            pass

        # popupframe.grid(row=0, columnspan=2)
        # B1.grid(row=1, column=0, sticky=W)
        # B2.grid(row=1, column=1, sticky=W)
        # self.main_frm.mainloop()



import os

os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE"
from typing import Dict, List, Optional, Union

try:
    from typing import Literal
except:
    from typing_extensions import Literal
try:
    from ultralytics import YOLO
except ModuleNotFoundError:
    YOLO = None

import numpy as np
import pandas as pd
import torch

from simba.data_processors.cuda.utils import _is_cuda_available
from simba.third_party_label_appenders.converters import \
    yolo_obb_data_to_bounding_box
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_instance, check_int,
                                check_str, check_valid_boolean,
                                check_valid_lst, get_fn_ext)
from simba.utils.data import df_smoother, savgol_smoother
from simba.utils.errors import SimBAGPUError, SimBAPAckageVersionError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import find_core_cnt, get_video_meta_data
from simba.utils.yolo import (_get_undetected_obs, check_valid_device,
                              load_yolo_model, yolo_predict)

COORD_COLS = ['X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']
OUT_COLS = ['FRAME', 'CLASS_ID', 'CLASS_NAME', 'CONFIDENCE', 'X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']
SMOOTHING_METHODS = ('savitzky-golay', 'bartlett', 'blackman', 'boxcar', 'cosine', 'gaussian', 'hamming', 'exponential')
NEAREST = 'nearest'

class YoloInference():
    """
    Performs object detection inference on a video using a YOLO model.

    This class performs YOLO-based object detection on one or more video files. It supports GPU acceleration,
    batch processing, streaming, and optional result saving. The model returns bounding box coordinates and
    class confidence scores for each frame. Results can be smoothed or interpolated to handle detection gaps.

    .. seealso::
       To perform bounding box and keypoint (pose) detection, see :func:`~simba.bounding_box_tools.yolo.yolo_pose_inference.YOLOPoseInference`

    :param Union[str, os.PathLike] weights: Path to the YOLO model weights file or a loaded YOLO model object.
    :param Union[str, os.PathLike] or List[Union[str, os.PathLike]] video_path: Path(s) to the input video file(s) for performing inference.
    :param Optional[bool] verbose: If True, outputs progress information and timing. Defaults to False.
    :param Optional[Union[str, os.PathLike]] save_dir: Directory to save the inference results (e.g., as CSV). If None, results are returned. Defaults to None.
    :param Optional[bool] half_precision: If True, uses half precision (fp16) for inference. Reduces memory usage and speeds up inference on supported devices. Defaults to True.
    :param Union[Literal['cpu'], int] device: Device to use for inference. Use 'cpu' for CPU or GPU index (e.g., 0 for CUDA:0). Defaults to 0.
    :param Optional[int] batch_size: Number of frames to process in parallel. Defaults to 4.
    :param int core_cnt: Number of CPU cores to use for preprocessing or parallel work. Defaults to 8.
    :param float threshold: Confidence threshold for object detections. Detections below this value are discarded. Defaults to 0.25.
    :param int max_detections: Maximum number of detections per frame. Defaults to 300.
    :param Optional[Literal['savitzky-golay', 'bartlett', 'blackman', 'boxcar', 'cosine', 'gaussian', 'hamming', 'exponential']] smoothing_method: Optional smoothing method to apply to bounding box sequences over time to reduce jitter.
    :param Optional[int] smoothing_time_window: Number of frames to consider for temporal smoothing. Effective only if smoothing_method is specified.
    :param bool interpolate: If True, interpolates missing bounding boxes using the 'nearest' method. Useful for filling detection gaps.
    :param int imgsz: Input image size for inference. Must be a square value. Defaults to 640.
    :param Optional[bool] stream: If True, performs inference using a generator that yields predictions iteratively, reducing memory usage. Defaults to True.
    :return: If `save_dir` is None, returns a dictionary of results keyed by video file names. Otherwise, results are saved to `save_dir`.
    :rtype: Union[None, Dict[str, pd.DataFrame]]

    :example:
    >>> video_path = "/mnt/d/netholabs/yolo_videos/input/mp4_20250606083508/2025-05-28_19-50-23.mp4"
    >>> i = YoloInference(weights=r"/mnt/c/troubleshooting/coco_data/mdl/train8/weights/best.pt", video_path=video_path, save_dir=r"/mnt/c/troubleshooting/coco_data/mdl/results", verbose=True, gpu=True)
    >>> i.run()
    """
    def __init__(self,
                 weights: Union[str, os.PathLike, YOLO],
                 video_path: Union[Union[str, os.PathLike], List[Union[str, os.PathLike]]],
                 verbose: Optional[bool] = False,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 half_precision: Optional[bool] = True,
                 device: Union[Literal['cpu'], int] = 0,
                 batch_size: Optional[int] = 4,
                 core_cnt: int = 8,
                 threshold: float = 0.25,
                 max_detections: int = 300,
                 smoothing_method: Optional[Literal['savitzky-golay', 'bartlett', 'blackman', 'boxcar', 'cosine', 'gaussian', 'hamming', 'exponential']] = None,
                 smoothing_time_window: Optional[int] = None,
                 interpolate: bool = False,
                 imgsz: int = 640,
                 stream: Optional[bool] = True) -> Union[None, Dict[str, pd.DataFrame]]:

        if not _is_cuda_available()[0]:
            raise SimBAGPUError(msg='No GPU detected.', source=self.__class__.__name__)
        if YOLO is None:
            raise SimBAPAckageVersionError(msg='ultralytics.YOLO package not detected.', source=self.__class__.__name__)
        if isinstance(video_path, list):
            check_valid_lst(data=video_path, source=f'{self.__class__.__name__} video_path', valid_dtypes=(str, np.str_,), min_len=1)
        elif isinstance(video_path, str):
            check_file_exist_and_readable(file_path=video_path)
            video_path = [video_path]
        for i in video_path:
            _ = get_video_meta_data(video_path=i)

        check_instance(source=f'{self.__class__.__name__} weights', instance=weights, accepted_types=(str, os.PathLike, YOLO))
        if not isinstance(weights, YOLO):
            check_file_exist_and_readable(file_path=weights)
            self.model = load_yolo_model(weights_path=weights, verbose=verbose, device=device)
        else:
            self.model = weights
        check_valid_boolean(value=[half_precision, verbose, stream, interpolate], source=self.__class__.__name__)
        check_int(name=f'{self.__class__.__name__} batch_size', value=batch_size, min_value=1)
        check_int(name=f'{self.__class__.__name__} imgsz', value=imgsz, min_value=1)
        check_int(name=f'{self.__class__.__name__} imgsz', value=imgsz, min_value=1)
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=0.0, max_value=1.0)
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=1, max_value=find_core_cnt()[0])
        check_valid_device(device=device)
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} save_dir')
        if smoothing_method is not None:
            check_str(name=f'{self.__class__.__name__} smoothing', value=smoothing_method, options=SMOOTHING_METHODS)
            check_float(name=f'{self.__class__.__name__} smoothing_time_window', value=smoothing_time_window, min_value=10e-6)
        torch.set_num_threads(core_cnt)
        self.video_path, self.half_precision, self.stream, self.batch_size = video_path, half_precision, stream, batch_size
        self.interpolate, self.smoothing_method, self.smoothing_time_window = interpolate, smoothing_method, smoothing_time_window
        self.save_dir, self.verbose, self.imgsz, self.core_cnt, self.device = save_dir, verbose, imgsz, core_cnt,device
        self.threshold, self.max_detections = threshold, max_detections

    def run(self):
        class_dict = self.model.names
        results = {}
        timer = SimbaTimer(start=True)
        for path in self.video_path:
            _, video_name, _ = get_fn_ext(filepath=path)
            video_meta_data = get_video_meta_data(video_path=path)
            video_out = []
            video_predictions = yolo_predict(model=self.model, source=path, half=self.half_precision, batch_size=self.batch_size, stream=self.stream, imgsz=self.imgsz, device=self.device, threshold=self.threshold, max_detections=self.max_detections, verbose=self.verbose)
            for frm_cnt, video_prediction in enumerate(video_predictions):
                boxes = video_prediction.obb.data if video_prediction.obb is not None else video_prediction.boxes.data
                boxes = boxes.cpu().numpy().astype(np.float32)
                detected_classes = np.unique(boxes[:, -1]).astype(int) if boxes.size > 0 else []
                for class_id, class_name in class_dict.items():
                    if class_id not in detected_classes:
                        video_out.append(_get_undetected_obs(frm_id=frm_cnt, class_id=class_id, class_name=class_name, value_cnt=9))
                        continue
                    cls_data = boxes[np.argwhere(boxes[:, -1] == class_id)]
                    cls_data = cls_data.reshape(-1, boxes.shape[1])[np.argmax(boxes[:, -2].flatten())]
                    if video_prediction.obb is not None:
                        box = yolo_obb_data_to_bounding_box(center_x=cls_data[0], center_y=cls_data[1], width=cls_data[2], height=cls_data[3], angle=cls_data[4]).flatten()
                    else:
                        box = np.array([cls_data[0], cls_data[1], cls_data[2], cls_data[1], cls_data[2], cls_data[3], cls_data[0], cls_data[3]]).astype(np.int32)
                    video_out.append([frm_cnt, cls_data[-1], class_dict[cls_data[-1]], cls_data[-2]] + list(box))
            results[video_name] = pd.DataFrame(video_out, columns=OUT_COLS)
            if self.interpolate:
                for cord_col in COORD_COLS:
                    results[video_name][cord_col] = results[video_name][cord_col].astype(np.float32).astype(np.int32).replace(to_replace=-1, value=np.nan)
                    results[video_name][cord_col] = results[video_name][cord_col].interpolate(method=NEAREST, axis=0).ffill().bfill()
            if self.smoothing_method is not None:
                if self.smoothing_method != 'savitzky-golay':
                    smoothened = df_smoother(data=results[video_name][COORD_COLS], fps=video_meta_data['fps'], time_window=self.smoothing_time_window, source=self.__class__.__name__, method=self.smoothing_method)
                else:
                    smoothened = savgol_smoother(data=results[video_name][COORD_COLS], fps=video_meta_data['fps'], time_window=self.smoothing_time_window, source=self.__class__.__name__)
                results[video_name].update(smoothened)
        timer.stop_timer()
        if not self.save_dir:
            if self.verbose:
                stdout_success(f'YOLO results created', elapsed_time=timer.elapsed_time_str)
            return results
        else:
            for k, v in results.items():
                save_path = os.path.join(self.save_dir, f'{k}.csv')
                v.to_csv(save_path)
            if self.verbose:
                stdout_success(f'YOLO results saved in {self.save_dir} directory', elapsed_time=timer.elapsed_time_str)

# video_path = r"/mnt/c/troubleshooting/mitra/project_folder/videos/501_MA142_Gi_CNO_0521.mp4"
# # video_path = "/mnt/d/netholabs/yolo_videos/input/mp4_20250606083508/2025-05-28_19-50-23.mp4"
# # video_path = r"D:\cvat_annotations\videos\s24-eating.mp4"
# weights_path = r"D:\cvat_annotations\yolo_07032025\bbox_annot\mdl\train10\weights\best.pt"
# save_dir = r"D:\cvat_annotations\yolo_07032025\out_data"
# i = YoloInference(weights_path=weights_path,
#                   video_path=video_path,
#                   save_dir=save_dir,
#                   stream=True,
#                   verbose=True)
# i.run()


import json
import os
import random
from typing import Optional, Union

import cv2
import h5py
import numpy as np

from simba.third_party_label_appenders.transform.utils import \
    create_yolo_keypoint_yaml
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_str,
                                check_valid_boolean)
from simba.utils.errors import InvalidInputError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (create_directory,
                                    find_files_of_filetypes_in_directory,
                                    find_video_of_file, get_fn_ext,
                                    read_frm_of_video, read_img)
from simba.utils.yolo import keypoint_array_to_yolo_annotation_str


class SleapAnnotations2Yolo():

    """
    Convert SLEAP annotations to YOLO formatted training data.

    :param Union[str, os.PathLike] data_dir: Directory containing SLEAP annotations `.slp` files
    :param Union[str, os.PathLike] save_dir: Directory to save YOLO-formatted images, labels, and metadata.
    :param bool verbose: If True, print progress during processing.
    :param float train_size: Proportion of frames to assign to the training set (rest go to validation).
    :param bool greyscale: If True, convert images to grayscale.
    :param bool clahe: If True, apply CLAHE (Contrast Limited Adaptive Histogram Equalization).
    :param float padding: Relative padding to apply around the bounding box of keypoints (range 0.0 to 1.0).
    :param Optional[str] single_id: Optional custom ID to assign all annotations the same class (used in single-animal datasets).

    :example:
    >>> runner = SleapAnnotations2Yolo(sleap_dir=r'D:\cvat_annotations\frames\slp_to_yolo', save_dir=r'D:\cvat_annotations\frames\slp_to_yolo\yolo')
    >>> runner.run()
    """


    def __init__(self,
                 sleap_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 video_dir: Optional[Union[str, os.PathLike]] = None,
                 padding: Optional[float] = None,
                 train_size: float = 0.8,
                 verbose: bool = True,
                 greyscale: bool = False,
                 clahe: bool = False,
                 single_id: Optional[str] = None):

        check_if_dir_exists(in_dir=os.path.dirname(save_dir), source=f'{self.__class__.__name__} save_path')
        check_if_dir_exists(in_dir=os.path.dirname(sleap_dir), source=f'{self.__class__.__name__} sleap_dir')
        check_float(name=f'{self.__class__.__name__} train_size', value=train_size, max_value=0.99, min_value=0.1)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale', raise_error=True)
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe', raise_error=True)
        if single_id is not None: check_str(name=f'{self.__class__.__name__} single_id', value=single_id, raise_error=True)
        if video_dir is not None: check_if_dir_exists(in_dir=video_dir)
        self.slp_paths = find_files_of_filetypes_in_directory(directory=sleap_dir, extensions=['.slp'], as_dict=True, raise_error=True)
        self.padding, self.train_size, self.verbose, self.save_dir = padding, train_size, verbose, save_dir
        self.clahe, self.greyscale = clahe, greyscale
        self.img_dir, self.lbl_dir = os.path.join(save_dir, 'images'), os.path.join(save_dir, 'labels')
        self.img_train_dir, self.img_val_dir = os.path.join(self.img_dir, 'train'), os.path.join(self.img_dir, 'val')
        self.lbl_train_dir, self.lb_val_dir = os.path.join(self.lbl_dir, 'train'), os.path.join(self.lbl_dir, 'val')
        create_directory(paths=[self.img_train_dir, self.img_val_dir, self.lbl_train_dir, self.lb_val_dir], overwrite=False)
        self.map_path = os.path.join(save_dir, 'map.yaml')
        self.single_id, self.video_dir = single_id, video_dir

    def __h5_to_dict(self, name_or_group):
        if isinstance(name_or_group, h5py.Dataset):
            return name_or_group[()]
        elif isinstance(name_or_group, h5py.Group):
            return {k: self.__h5_to_dict(name_or_group[k]) for k in name_or_group}
        else:
            return None

    def run(self):
        timer = SimbaTimer(start=True)
        lbl_cnt, sleap_data = 0, {}
        for file_cnt, (file_name, file_path) in enumerate(self.slp_paths.items()):
            with h5py.File(file_path, "r") as f:
                data = self.__h5_to_dict(f)
            check_if_keys_exist_in_dict(data=data, key=['frames', 'instances', 'points', 'videos_json'])
            img_names = json.loads(data["videos_json"].item().decode("utf-8"))

            if not 'backend' in img_names.keys() and not 'filename' in img_names.keys():
                raise InvalidInputError(msg='Could not find backend or filename keys in videos_json', source=self.__class__.__name__)
            #check_if_keys_exist_in_dict(data=img_names, key=['filename'])
            sleap_data[file_name] = data

        for file_cnt, (file_name, data) in enumerate(sleap_data.items()):
            lbl_cnt += len(data["frames"])

        train_idx = random.sample(list(range(0, lbl_cnt)), int(lbl_cnt * self.train_size))
        img_cnt, kpt_shapes, names = 0, set(), {}
        for file_cnt, (file_name, file_path) in enumerate(self.slp_paths.items()):
            with h5py.File(file_path, "r") as f:
                data = self.__h5_to_dict(f)
            frames = data["frames"]
            instances = data["instances"]
            points = data["points"]
            img_names, video_path = json.loads(data["videos_json"].item().decode("utf-8")), None
            if 'backend' in img_names.keys():
                video_path = img_names['backend']['filename']
                if not os.path.isfile(video_path):
                    if self.video_dir is None:
                        raise NoFilesFoundError(msg=f'Could not locate file {video_path}, pass a video directory', source=self.__class__.__name__)
                    else:
                        video_path = find_video_of_file(video_dir=self.video_dir, filename=get_fn_ext(filepath=video_path)[1], raise_error=True)
            for i, frame_data in enumerate(frames):
                if self.verbose: print(f'Converting SLEAP annotation to YOLO {img_cnt+1}/{lbl_cnt}...')
                frame_idx = frame_data[0]
                if video_path is None:
                    frm_path = img_names['filename'][frame_idx]
                    file_name = get_fn_ext(filepath=frm_path)[1]
                    check_file_exist_and_readable(file_path=frm_path, raise_error=True)
                    img = read_img(img_path=frm_path, greyscale=self.greyscale, clahe=self.clahe)
                else:
                    img = read_frm_of_video(video_path=video_path, frame_index=frame_idx, greyscale=self.greyscale, clahe=self.clahe)
                img_w, img_h = img.shape[1], img.shape[0]
                frm_instances = [inst for inst in instances if int(inst[2]) == frame_idx]
                if img_cnt in train_idx:
                    img_save_path, lbl_save_path = os.path.join(self.img_train_dir, f'{file_name}.png'), os.path.join(self.lbl_train_dir, f'{file_name}.txt')
                else:
                    img_save_path, lbl_save_path = os.path.join(self.img_val_dir, f'{file_name}.png'), os.path.join(self.lb_val_dir, f'{file_name}.txt')
                img_lbl = ''
                for instance in frm_instances:

                    if instance[1] not in names.keys():
                        names[int(instance[1])] = f'animal_{int(instance[1])}'
                    instance_str = f'{instance[1]} ' if self.single_id is None else '0 '
                    pt_start = int(instance[7])
                    pt_end = int(instance[8])
                    pts = points[pt_start:pt_end]
                    keypoints = np.array([[p[0], p[1]] if p[2] else [0.0, 0.0] for p in pts]).astype(np.int32)
                    if keypoints.shape[0] == 0 or np.all(keypoints == 0):
                        continue
                    keypoints = np.nan_to_num(keypoints, nan=0.0)
                    kpt_shapes.add(keypoints.shape[0])
                    visability_col = np.full((keypoints.shape[0], 1), fill_value=2).flatten()
                    keypoints = np.insert(keypoints, 2, visability_col, axis=1)
                    both_zero = (keypoints[:, 0] == 0) & (keypoints[:, 1] == 0)
                    has_nan_or_inf = ~np.isfinite(keypoints[:, 0]) | ~np.isfinite(keypoints[:, 1])
                    mask = both_zero | has_nan_or_inf
                    keypoints[mask, 2] = 0
                    keypoints[~np.isfinite(keypoints)] = 0
                    instance_str += keypoint_array_to_yolo_annotation_str(x=keypoints, img_w=img_w, img_h=img_h, padding=self.padding)
                    img_lbl += instance_str
                with open(lbl_save_path, mode='wt', encoding='utf-8') as f:
                    f.write(img_lbl)
                cv2.imwrite(img_save_path, img)
                img_cnt+=1

        if len(list(kpt_shapes)) > 1:
            raise InvalidInputError(msg=f'Found more than 1 keypoint shapes: {kpt_shapes}', source=self.__class__.__name__)

        names = names if self.single_id is None else {0: self.single_id}
        create_yolo_keypoint_yaml(path=self.save_dir, train_path=self.img_train_dir, val_path=self.img_val_dir, names=names, save_path=self.map_path, kpt_shape=(list(kpt_shapes)[0], 3), flip_idx=tuple(list(range(0, list(kpt_shapes)[0]))))
        timer.stop_timer()
        if self.verbose:
            stdout_success(msg=f'YOLO annotations saved at {self.save_dir}...', source=self.__class__.__name__, elapsed_time=timer.elapsed_time_str)

# runner = SleapAnnotations2Yolo(sleap_dir=r'D:\slp_predictions', save_dir=r'D:\slp_predictions\yolo', single_id='mouse')
# runner.run()






__author__ = "Simon Nilsson"

import os
from datetime import datetime
from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.transform.labelme_to_df import \
    LabelMe2DataFrame
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimBADropDown)
from simba.utils.checks import check_if_dir_exists
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    str_2_bool)


class Labelme2DataFramePopUp(PopUpMixin):

    """
    :example:
    >>> Labelme2DataFramePopUp()
    """

    def __init__(self):
        PopUpMixin.__init__(self, title="LABELME JSONS TO DATAFRAME", icon='labelme')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.labelme_dir = FolderSelect(settings_frm, folderDescription="LABELME DIRECTORY (JSONS):", lblwidth=35, entry_width=40, initialdir=r"C:\troubleshooting\coco_data\labels\test_2",)
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY :", lblwidth=35, entry_width=40, initialdir=r"C:\troubleshooting\RAT_NOR")
        self.grey_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="GREYSCALE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.clahe_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="CLAHE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE: ", label_width=35, dropdown_width=40, value='TRUE')
        self.pad_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="PAD: ", label_width=35, dropdown_width=40, value='FALSE')
        self.normalize_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="NORMALIZE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.size_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['MIN', 'MAX', 'NONE'], label="SIZE: ", label_width=35, dropdown_width=40, value='MAX')

        settings_frm.grid(row=0, column=0, sticky=NW)
        self.labelme_dir .grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.grey_dropdown.grid(row=2, column=0, sticky=NW)
        self.clahe_dropdown.grid(row=3, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=4, column=0, sticky=NW)
        self.pad_dropdown.grid(row=5, column=0, sticky=NW)
        self.normalize_dropdown.grid(row=6, column=0, sticky=NW)
        self.size_dropdown.grid(row=7, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()


    def run(self):
        labelme_dir = self.labelme_dir.folder_path
        save_dir = self.save_dir.folder_path
        grey = str_2_bool(self.grey_dropdown.get_value())
        clahe = str_2_bool(self.clahe_dropdown.get_value())
        verbose = str_2_bool(self.verbose_dropdown.get_value())
        pad = str_2_bool(self.pad_dropdown.get_value())
        normalize = str_2_bool(self.normalize_dropdown.get_value())
        size = self.size_dropdown.get_value()
        size = None if size == 'NONE' else size.lower()
        _ = find_files_of_filetypes_in_directory(directory=labelme_dir, extensions=['.json'], raise_error=True)
        check_if_dir_exists(in_dir=labelme_dir, source=f'{self.__class__.__name__} LABELME DIRECTORY (JSONS)', raise_error=True)
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} SAVE DIRECTORY', raise_error=True)

        save_path = os.path.join(save_dir, f'labelme_dataframe_{datetime.now().strftime("%Y%m%d%H%M%S")}.csv')
        runner = LabelMe2DataFrame(labelme_dir=labelme_dir, greyscale=grey, pad=pad, size=size, normalize=normalize, save_path=save_path, verbose=verbose, clahe=clahe)
        runner.run()


#Labelme2DataFramePopUp()


__author__ = "Simon Nilsson"

import os
import threading
from tkinter import *
from typing import Union

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.single_run_model_validation_video import \
    ValidateModelOneVideo
from simba.plotting.single_run_model_validation_video_mp import \
    ValidateModelOneVideoMultiprocess
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, SimBADropDown
from simba.utils.checks import check_float, check_int
from simba.utils.enums import Links, Options
from simba.utils.errors import InvalidInputError, NoFilesFoundError
from simba.utils.read_write import str_2_bool

AUTO = 'AUTO'
GANTT_FRAME = "Gantt chart: final frame only (slightly faster)"
GANTT_VIDEO = "Gantt chart: video"

FONT_SIZE_OPTIONS = list(range(1, 56, 1))
FONT_SIZE_OPTIONS.insert(0, AUTO)
TEXT_SPACE_OPTIONS = list(range(1, 106, 2))
TEXT_SPACE_OPTIONS.insert(0, AUTO)

OPACITY_OPTIONS = list(np.arange(0.1, 1.1, 0.1))
OPACITY_OPTIONS = [round(x, 1) for x in OPACITY_OPTIONS]
POSE_PALETTE_OPTIONS = Options.PALETTE_OPTIONS_CATEGORICAL.value + Options.PALETTE_OPTIONS.value

print(POSE_PALETTE_OPTIONS)

class ValidationVideoPopUp(PopUpMixin, ConfigReader):
    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 feature_path: Union[str, os.PathLike],
                 model_path: Union[str, os.PathLike],
                 discrimination_threshold: float,
                 shortest_bout: int):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if not os.path.isfile(feature_path):
            raise NoFilesFoundError(msg=f'Set DATA FEATURE FILE PATH to a valid file before creating validation video: Got {model_path}.', source=self.__class__.__name__)
        if not os.path.isfile(model_path):
            raise NoFilesFoundError(msg=f'Set MODEL FILE PATH to a valid file before creating validation video: Got {model_path}.', source=self.__class__.__name__)
        if not check_float(name=f'{self.__class__.__name__} discrimination_threshold', min_value=0.0, max_value=1.0, value=discrimination_threshold, raise_error=False):
            raise InvalidInputError(msg=f'Set the DISCRIMINATION THRESHOLD to a value between 0.0 and 1.0 before creating validation video. Got: {discrimination_threshold}.', source=self.__class__.__name__)
        print(discrimination_threshold)
        discrimination_threshold = float(discrimination_threshold)
        if not check_int(name=f'{self.__class__.__name__} shortest_bout', value=shortest_bout, min_value=0, raise_error=False):
            raise InvalidInputError(msg=f'Set the MINIMUM BOUT LENGTH (MS) to a value above 1 before creating validation video. Got {shortest_bout}.', source=self.__class__.__name__)
        shortest_bout = int(shortest_bout)
        self.mdl_path, self.feature_path = model_path, feature_path
        PopUpMixin.__init__(self, title="CREATE VALIDATION VIDEO", icon='tick')
        self.discrimination_threshold, self.shortest_bout = float(discrimination_threshold), int(shortest_bout)
        style_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="STYLE SETTINGS", icon_name='style', icon_link=Links.OUT_OF_SAMPLE_VALIDATION.value, padx=5, pady=5, relief='solid')
        self.font_size_dropdown = SimBADropDown(parent=style_frm, dropdown_options=FONT_SIZE_OPTIONS, label='FONT SIZE: ', label_width=30, dropdown_width=40, value=AUTO)
        self.space_dropdown = SimBADropDown(parent=style_frm, dropdown_options=TEXT_SPACE_OPTIONS, label='TEXT SPACING: ', label_width=30, dropdown_width=40, value=AUTO)
        self.circle_size_dropdown = SimBADropDown(parent=style_frm, dropdown_options=TEXT_SPACE_OPTIONS, label='CIRCLE SIZE: ', label_width=30, dropdown_width=40, value=AUTO)
        self.text_opacity_dropdown = SimBADropDown(parent=style_frm, dropdown_options=OPACITY_OPTIONS, label='TEXT OPACITY: ', label_width=30, dropdown_width=40, value=0.8)
        self.text_thickness_dropdown = SimBADropDown(parent=style_frm, dropdown_options=FONT_SIZE_OPTIONS, label='TEXT THICKNESS: ', label_width=30, dropdown_width=40, value=2)
        self.bp_palette_dropdown = SimBADropDown(parent=style_frm, dropdown_options=POSE_PALETTE_OPTIONS, label='BODY-PART PALETTE: ', label_width=30, dropdown_width=40, value=2)

        style_frm.grid(row=0, column=0, sticky=NW, padx=10, pady=10)
        self.font_size_dropdown.grid(row=0, column=0, sticky=NW)
        self.space_dropdown.grid(row=1, column=0, sticky=NW)
        self.circle_size_dropdown.grid(row=2, column=0, sticky=NW)
        self.text_opacity_dropdown.grid(row=3, column=0, sticky=NW)
        self.text_thickness_dropdown.grid(row=4, column=0, sticky=NW)
        self.bp_palette_dropdown.grid(row=5, column=0, sticky=NW)

        tracking_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="TRACKING OPTIONS", icon_name='pose', icon_link=Links.OUT_OF_SAMPLE_VALIDATION.value, padx=5, pady=5, relief='solid')
        self.show_pose_dropdown = SimBADropDown(parent=tracking_frm, dropdown_options=['TRUE', 'FALSE'], label='SHOW POSE: ', label_width=30, dropdown_width=40, value='TRUE')
        self.show_animal_names_dropdown = SimBADropDown(parent=tracking_frm, dropdown_options=['TRUE', 'FALSE'], label='SHOW ANIMAL NAMES: ', label_width=30, dropdown_width=40, value='FALSE')
        self.core_cnt_dropdown = SimBADropDown(parent=tracking_frm, dropdown_options=list(range(1, self.cpu_cnt + 1)), label='CPU COUNT: ', label_width=30, dropdown_width=40, value=int(self.cpu_cnt/2))

        tracking_frm.grid(row=1, column=0, sticky=NW, padx=10, pady=10)
        self.show_pose_dropdown.grid(row=0, column=0, sticky=NW)
        self.show_animal_names_dropdown.grid(row=1, column=0, sticky=NW)
        self.core_cnt_dropdown.grid(row=2, column=0, sticky=NW)

        gantt_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="GANTT SETTINGS", icon_name='gantt_small', icon_link=Links.OUT_OF_SAMPLE_VALIDATION.value, padx=5, pady=5, relief='solid')
        self.gantt_dropdown = SimBADropDown(parent=gantt_frm, dropdown_options= Options.GANTT_VALIDATION_OPTIONS.value, label='GANTT TYPE:', label_width=30, dropdown_width=40, value=Options.GANTT_VALIDATION_OPTIONS.value[0])
        gantt_frm.grid(row=3, column=0, sticky=NW, padx=10, pady=10)
        self.gantt_dropdown.grid(row=0, column=0, sticky=NW)


        self.create_run_frm(run_function=self.__run)
        self.main_frm.mainloop()

    def __run(self):
        show_pose = str_2_bool(self.show_pose_dropdown.getChoices())
        show_animal_names = str_2_bool(self.show_animal_names_dropdown.getChoices())
        print(self.font_size_dropdown.getChoices())
        font_size = None if self.font_size_dropdown.getChoices() == AUTO else int(self.font_size_dropdown.getChoices())
        circle_size = None if self.circle_size_dropdown.getChoices() == AUTO else int(self.circle_size_dropdown.getChoices())
        text_space_scale = None if self.space_dropdown.getChoices() == AUTO else int(self.space_dropdown.getChoices())
        text_thickness = int(self.text_thickness_dropdown.getChoices())
        text_opacity = float(self.text_opacity_dropdown.getChoices())
        core_cnt = int(self.core_cnt_dropdown.getChoices())
        create_gantt = self.gantt_dropdown.getChoices()
        if create_gantt.strip() == GANTT_FRAME: create_gantt = 1
        elif create_gantt.strip() == GANTT_VIDEO: create_gantt = 2
        else: create_gantt = None

        if core_cnt == 1:
            validation_video_creator = ValidateModelOneVideo(config_path=self.config_path,
                                                             feature_path=self.feature_path,
                                                             model_path=self.mdl_path,
                                                             discrimination_threshold=self.discrimination_threshold,
                                                             shortest_bout=self.shortest_bout,
                                                             font_size=font_size,
                                                             create_gantt=create_gantt,
                                                             circle_size=circle_size,
                                                             text_spacing=text_space_scale,
                                                             show_pose=show_pose,
                                                             text_thickness=text_thickness,
                                                             text_opacity=text_opacity,
                                                             show_animal_names=show_animal_names)


        else:
            validation_video_creator = ValidateModelOneVideoMultiprocess(config_path=self.config_path,
                                                                         feature_path=self.feature_path,
                                                                         model_path=self.mdl_path,
                                                                         discrimination_threshold=self.discrimination_threshold,
                                                                         shortest_bout=self.shortest_bout,
                                                                         font_size=font_size,
                                                                         create_gantt=create_gantt,
                                                                         circle_size=circle_size,
                                                                         text_spacing=text_space_scale,
                                                                         show_pose=show_pose,
                                                                         text_thickness=text_thickness,
                                                                         text_opacity=text_opacity,
                                                                         show_animal_names=show_animal_names,
                                                                         core_cnt=core_cnt)

        threading.Thread(target=validation_video_creator.run()).start()



# ValidationVideoPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                      feature_path=r"C:\troubleshooting\mitra\project_folder\csv\features_extracted\501_MA142_Gi_CNO_0521.csv",
#                      model_path=r"C:\troubleshooting\mitra\models\generated_models\grroming_undersample_2_1000\grooming.sav",
#                      discrimination_threshold=0.4,
#                      shortest_bout=500)


import json
import os
from typing import Any, Dict, Optional, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.third_party_label_appenders.transform.utils import arr_to_b64
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_instance,
                                check_valid_boolean)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_img, recursive_file_search


class DLC2Labelme():

    """
    Convert a folder of DLC annotations into labelme json format.

    .. seealso::
       For Labelme -> DLC annotation conversion, see :func:`simba.third_party_label_appenders.transform.labelme_to_dlc.Labelme2DLC`

    :param Union[str, os.PathLike] dlc_dir: Folder with DLC annotations. I.e., directory inside
    :param Union[str, os.PathLike] save_dir: Directory to where to save the labelme json files.
    :param Optional[str] labelme_version: Version number encoded in the json files.
    :param Optional[Dict[Any, Any] flags: Flags included in the json files.
    :param Optional[bool] verbose: If True, prints progress.
    :return: None

    :example:
    >>> DLC2Labelme(dlc_dir="D:\TS_DLC\labeled-data\ts_annotations", save_dir="C:\troubleshooting\coco_data\labels\test").run()
    >>> DLC2Labelme(dlc_dir=r'D:\rat_resident_intruder\dlc_data\WIN_20190816081353', save_dir=r'D:\rat_resident_intruder\labelme').run()
    """

    def __init__(self,
                 dlc_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 labelme_version: Optional[str] = '5.3.1',
                 flags: Optional[Dict[Any, Any]] = None,
                 verbose: bool = True,
                 greyscale: bool = False,
                 clahe: bool = False) -> None:

        check_if_dir_exists(dlc_dir, source=f'{self.__class__.__name__} dlc_dir')
        check_if_dir_exists(save_dir, source=f'{self.__class__.__name__} save_dir')
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose')
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale')
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe')
        self.data_paths = recursive_file_search(directory=dlc_dir, substrings='CollectedData', extensions='csv', raise_error=True)
        self.version, self.verbose, self.save_dir = labelme_version, verbose, save_dir
        if flags is not None:
            check_instance(source=f'{self.__class__.__name__} flags', instance=flags, accepted_types=(dict,))
        self.flags = {} if flags is None else {}
        self.clahe, self.greyscale = clahe, greyscale

    def run(self):
        timer = SimbaTimer(start=True)
        self.body_parts_per_file, self.filecnt = {}, 0
        for file_cnt, file_path in enumerate(self.data_paths):
            file_dir = os.path.dirname(file_path)
            video_name = os.path.basename(os.path.dirname(file_path))
            body_part_headers = ['image']
            annotation_data = pd.read_csv(file_path, header=[0, 1, 2])
            body_parts = []
            for i in annotation_data.columns[1:]:
                if 'unnamed:' not in i[1].lower() and i[1] not in body_parts:
                    body_parts.append(i[1])
            for i in body_parts:
                body_part_headers.append(f'{i}_x'); body_part_headers.append(f'{i}_y')
            self.body_parts_per_file[file_path] = body_part_headers
            print(annotation_data)
            annotation_data.columns = body_part_headers
            for cnt, (idx, idx_data) in enumerate(annotation_data.iterrows()):
                if self.verbose:
                    print(f'Processing image {cnt + 1}/{len(annotation_data)}... (video {file_cnt + 1}/{len(self.data_paths)} ({video_name}))')
                _, img_name, ext = get_fn_ext(filepath=idx_data['image'])
                video_img_name = f'{video_name}.{img_name}'
                img_path = os.path.join(file_dir, os.path.join(f'{img_name}{ext}'))
                check_file_exist_and_readable(file_path=img_path)
                img = read_img(img_path=img_path, clahe=self.clahe, greyscale=self.greyscale)
                idx_data = idx_data.to_dict()
                shapes = []
                for bp_name in body_parts:
                    img_shapes = {'label': bp_name,
                                  'points': [idx_data[f'{bp_name}_x'], idx_data[f'{bp_name}_y']],
                                  'group_id': None,
                                  'description': "",
                                  'shape_type': 'point',
                                  'flags': {}}
                    shapes.append(img_shapes)
                out = {"version": self.version,
                       'flags': self.flags,
                       'shapes': shapes,
                       'imagePath': img_path,
                       'imageData': arr_to_b64(img),
                       'imageHeight': img.shape[0],
                       'imageWidth': img.shape[1]}
                save_path = os.path.join(self.save_dir, f'{video_img_name}.json')
                with open(save_path, "w") as f:
                    json.dump(out, f)
                self.filecnt += 1
        timer.stop_timer()
        if self.verbose:
            stdout_success(f'Labelme data for {self.filecnt} image(s) saved in {self.save_dir} directory', elapsed_time=timer.elapsed_time_str)


#DLC2Labelme(dlc_dir=r'D:\rat_resident_intruder\dlc_data\WIN_20190816081353', save_dir=r'D:\rat_resident_intruder\labelme').run()

# DLC_DIR = r"D:\TS_DLC\labeled-data\ts_annotations"
# SAVE_DIR = r"C:\troubleshooting\coco_data\labels\test"
#
# runner = DLC2Labelme(dlc_dir=DLC_DIR, save_dir=SAVE_DIR)
# runner.run()
#


    #>>> DLC2Labelme(dlc_dir=r'D:\rat_resident_intruder\dlc_data\WIN_20190816081353', save_dir=r'D:\rat_resident_intruder\labelme').run()